#include "StdAfx.h"
#include "Action.h"
#include "Error.h"


NameToMessage_map_t		Action::map_NameToMessage;
NameToVirtualKey_map_t	Action::map_NameToVirtualKey;
NameToAction_map_t		Action::map_NameToAction;
ExtendedKey_set_t		Action::set_ExtendedKey;
INPUT*					Action::p_InputQueue = NULL;
long					Action::l_QueueLength = 0;


//
bool SubAction::Act (UIObject* _pUIObject, CWnd* _pWnd, CPoint& _rPos, const char* _zText)
{
	//
	if ((e_at_SendMessage == e_ActionType) &&
		(NULL != _pUIObject) &&
		(NULL != _pWnd))
	{
		NameToMessage_map_t::iterator	ite;
		ite = Action::map_NameToMessage.find (s_MessageName);
		if (ite == Action::map_NameToMessage.end ())
		{
			String sError;
			sError << "Unknown message name [" << s_MessageName << "].";
			Error::AddError (sError);
			return false;
		}
		UINT uiMessage = ite->second;
		WPARAM wParam = (WPARAM)(long) s_WParam;
		if ("[command_id]" == s_WParam)
			wParam = (WPARAM) _pUIObject->i_CommandID;
		else
		{
			ite = Action::map_NameToMessage.find (s_WParam);
			if (ite != Action::map_NameToMessage.end ())
				wParam = ite->second;
		}
		LPARAM lParam = (LPARAM)(long) s_LParam;

		_pWnd->SendMessage (uiMessage, wParam, lParam);
		return true;
	}

	//
	if ((e_at_PostMessage == e_ActionType) &&
		(NULL != _pUIObject) &&
		(NULL != _pWnd))

	{
		NameToMessage_map_t::iterator	ite;
		ite = Action::map_NameToMessage.find (s_MessageName);
		if (ite == Action::map_NameToMessage.end ())
		{
			String sError;
			sError << "Unknown message name [" << s_MessageName << "].";
			Error::AddError (sError);
			return false;
		}
		
		UINT uiMessage = ite->second;
		WPARAM wParam = (WPARAM)(long) s_WParam;
		if ("[command_id]" == s_WParam)
			wParam = (WPARAM) _pUIObject->i_CommandID;
		LPARAM lParam = (LPARAM)(long) s_LParam;

		_pWnd->PostMessage (uiMessage, wParam, lParam);
		return true;
	}

	//
	if (e_at_SendMouse == e_ActionType)
	{
		INPUT& input = Action::p_InputQueue [Action::l_QueueLength ++];
		
		input.type = INPUT_MOUSE;
		input.mi.dx = _rPos.x;
		input.mi.dy = _rPos.y;

		if ("left_down" == s_MouseFlags)
			input.mi.dwFlags |= MOUSEEVENTF_LEFTDOWN;
		if ("left_up" == s_MouseFlags)
			input.mi.dwFlags |= MOUSEEVENTF_LEFTUP;
	    if ("right_down" == s_MouseFlags)
			input.mi.dwFlags |= MOUSEEVENTF_RIGHTDOWN;
		if ("right_up" == s_MouseFlags)
			input.mi.dwFlags |= MOUSEEVENTF_RIGHTUP;
		if ("middle_down" == s_MouseFlags)
			input.mi.dwFlags |= MOUSEEVENTF_MIDDLEDOWN;
		if ("middle_up" == s_MouseFlags)
			input.mi.dwFlags |= MOUSEEVENTF_MIDDLEUP;
		if (true == s_MouseFlags.Has ("move"))
			input.mi.dwFlags |= MOUSEEVENTF_MOVE;
		if (true == s_MouseFlags.Has ("absolute_position"))
			input.mi.dwFlags |= MOUSEEVENTF_ABSOLUTE;

		return true;
	}

	//
	if (e_at_SendKey == e_ActionType)
	{
		NameToVirtualKey_map_t::iterator	ite;
		ite = Action::map_NameToVirtualKey.find (s_VirtualKey);
		if (ite == Action::map_NameToVirtualKey.end ())
		{
			String sError;
			sError << "Unknown virtual key name [" << s_VirtualKey << "].";
			Error::AddError (sError);
			return false;
		}

		INPUT& input = Action::p_InputQueue [Action::l_QueueLength ++];

		input.type = INPUT_KEYBOARD;
		input.ki.wVk = ite->second;
		input.ki.dwFlags = 0;
		if (Action::set_ExtendedKey.end () != Action::set_ExtendedKey.find (s_VirtualKey))
			input.ki.dwFlags |= KEYEVENTF_EXTENDEDKEY;
		if (true == s_KeyFlags.Has ("key_up"))
			input.ki.dwFlags |= KEYEVENTF_KEYUP;

		return true;
	}

	//
	if (e_at_TypeText == e_ActionType)
	{
		long lInputCount = 2 * strlen (_zText);
		INPUT* pInput = new INPUT [lInputCount];
		memset (pInput, 0, lInputCount * sizeof (INPUT));

		char* pCh = (char*)_zText;
		for (long i = 0; i < lInputCount; i += 2)
		{
			pInput [i].type = INPUT_KEYBOARD;
			pInput [i].ki.wScan = _T(*pCh);
			pInput [i].ki.wVk = 0;
			pInput [i].ki.dwFlags = KEYEVENTF_UNICODE;

			pInput [i+1].type = INPUT_KEYBOARD;
			pInput [i+1].ki.wScan = _T(*pCh);
			pInput [i+1].ki.wVk = 0;
			pInput [i+1].ki.dwFlags = KEYEVENTF_UNICODE|KEYEVENTF_KEYUP;

			++ pCh;
		}

		SendInput (lInputCount, pInput, sizeof (INPUT));
		delete[] pInput;
		return true;
	}
	return false;
}


//
Action::Action(void)
{
}

Action::~Action(void)
{
	ITERATE (SubAction_dq_t, dq_SubActions, ite)
		delete *ite;
	dq_SubActions.clear ();
}


//
bool Action::Act (UIObject* _pUIObject, CWnd* _pWnd, CPoint& _rPos, const char* _zText)
{
	if (NULL != p_InputQueue)
		delete[] p_InputQueue;
	long lMaxInputs = dq_SubActions.size ();
	p_InputQueue = new INPUT [lMaxInputs];
	memset (p_InputQueue, 0, lMaxInputs * sizeof (INPUT));
	l_QueueLength = 0;

	//
	bool bError = false;
	ITERATE (SubAction_dq_t, dq_SubActions, ite)
	{
		SubAction* pSubAction = *ite;
		if ((e_at_SendMouse != pSubAction->e_ActionType) &&
			(e_at_SendKey != pSubAction->e_ActionType))
		{
			if (l_QueueLength > 0)
			{
				long lSent = SendInput (l_QueueLength, p_InputQueue, sizeof (INPUT));
				memset (p_InputQueue, 0, lMaxInputs * sizeof (INPUT));
				l_QueueLength = 0;
			}
		}

		if (false == pSubAction->Act (_pUIObject, _pWnd, _rPos, _zText))
		{
			bError = true;
			break;
		}
	}

	//
	if (true == bError)
	{
		delete[] p_InputQueue;
		p_InputQueue = NULL;
		return false;
	}

	//
	if (l_QueueLength > 0)
	{
		long lSent = SendInput (l_QueueLength, p_InputQueue, sizeof (INPUT));
	}
	delete[] p_InputQueue;
	p_InputQueue = NULL;

	return true;
}


//
Action* Action::Find (const char* _zAction)
{
	NameToAction_map_t::iterator	ite;
	ite = Action::map_NameToAction.find (_zAction);
	if (ite == Action::map_NameToAction.end ())
	{
		String sError;
		sError << "Action::Find failed on '" << _zAction << "'.";
		Error::AddError (sError);
		return NULL;
	}
	return ite->second;
}


//
bool Action::LoadConfigs (void)
{
	LoadMessages ();
	LoadVirtualKeys ();
	LoadExtendedKeys ();
	return LoadActionConfigs ();
}


//
bool Action::LoadMessages (void)
{
	String_dq_t dqLines;
	if (false == File::ReadLines ("messages.cfg", dqLines))
	{
		Error::AddError ("Failed to open file 'messages.cfg'.");
		return false;
	}
	ITERATE (String_dq_t, dqLines, ite)
	{
		(*ite).Strip ();
		if ("" == *ite)
			continue;
		String_dq_t dqTokens;
		(*ite).Split (dqTokens);
		
		long lMessageID;
		sscanf_s (dqTokens [0], "%x", &lMessageID);
		Action::map_NameToMessage.insert (make_pair (dqTokens [1], (UINT)lMessageID));
	}
	return true;
}


//
bool Action::LoadVirtualKeys (void)
{
	String_dq_t dqLines;
	if (false == File::ReadLines ("virtualkeys.cfg", dqLines))
	{
		Error::AddError ("Failed to open file 'virtualkeys.cfg'.");
		return false;
	}
	ITERATE (String_dq_t, dqLines, ite)
	{
		(*ite).Strip ();
		if ("" == *ite)
			continue;
		String_dq_t dqTokens;
		(*ite).Split (dqTokens);

		long lKeyID;
		sscanf_s (dqTokens [0], "%x", &lKeyID);
		Action::map_NameToVirtualKey.insert (make_pair (dqTokens [1], (WORD)lKeyID));
	}
	return false;
}


//
bool Action::LoadExtendedKeys (void)
{
	String_dq_t dqLines;
	if (false == File::ReadLines ("extendedkeys.cfg", dqLines))
	{
		Error::AddError ("Failed to open file 'extendedkeys.cfg'.");
		return false;
	}
	long lLine = 0;
	ITERATE (String_dq_t, dqLines, ite)
	{
		++ lLine;
		(*ite).Strip ();
		if ("" == *ite)
			continue;

		if (Action::map_NameToVirtualKey.end () == Action::map_NameToVirtualKey.find (*ite))
		{
			String sError;
			sError << "[line " << lLine << "] Unknown virtual key (" << *ite << ") defined as extended key.";
			continue;
		}
		Action::set_ExtendedKey.insert (*ite);
	}
	return false;
}


//
bool Action::LoadActionConfigs (void)
{
	String_dq_t	dqLines;
	if (false == File::ReadLines ("actions.cfg", dqLines))
	{
		Error::AddError ("Failed to open file 'actions.cfg'.");
		return false;
	}

	Action* pAction = NULL;
	SubAction* pSubAction = NULL;
	int iLine = 0;
	ITERATE (String_dq_t, dqLines, ite)
	{
		++ iLine;
		String& rLine = *ite;
		rLine.Strip ();
		if ("" == rLine)
			continue;
		if ('#' == rLine.at (0))
			continue;

		if (true == rLine.Has ("action_type:"))
		{
			pSubAction = new SubAction;
			pAction->dq_SubActions.push_back (pSubAction);

			if (true == rLine.Has ("send message"))
				pSubAction->e_ActionType = e_at_SendMessage;
			else if (true == rLine.Has ("post message"))
				pSubAction->e_ActionType = e_at_PostMessage;
			else if (true == rLine.Has ("mouse"))
				pSubAction->e_ActionType = e_at_SendMouse;
			else if (true == rLine.Has ("key"))
				pSubAction->e_ActionType = e_at_SendKey;
			else if (true == rLine.Has ("type"))
				pSubAction->e_ActionType = e_at_TypeText;
			else
			{
				String sError;
				sError << "[line " << iLine << "] Unknown action type [" << rLine << "]";
				Error::AddError (sError);
			}
		}
		else if (true == rLine.Has ("message_name:"))
		{
			if ((e_at_SendMessage != pSubAction->e_ActionType) &&
				(e_at_PostMessage != pSubAction->e_ActionType))
			{
				String sError;
				sError << "[line " << iLine << "] message_name option found in non message action.";
				Error::AddError (sError);
			}
			rLine.Replace ("message_name:", "");
			rLine.Strip ();
			pSubAction->s_MessageName = rLine;
		}
		else if (true == rLine.Has ("wparam:"))
		{
			if ((e_at_SendMessage != pSubAction->e_ActionType) &&
				(e_at_PostMessage != pSubAction->e_ActionType))
			{
				String sError;
				sError << "[line " << iLine << "] wparam option found in non message action.";
				Error::AddError (sError);
			}
			rLine.Replace ("wparam:", "");
			rLine.Strip ();
			pSubAction->s_WParam = rLine;
		}
		else if (true == rLine.Has ("lparam:"))
		{
			if ((e_at_SendMessage != pSubAction->e_ActionType) &&
				(e_at_PostMessage != pSubAction->e_ActionType))
			{
				String sError;
				sError << "[line " << iLine << "] lparam option found in non message action.";
				Error::AddError (sError);
			}
			rLine.Replace ("lparam:", "");
			rLine.Strip ();
			pSubAction->s_LParam = rLine;
		}
		else if (true == rLine.Has ("mouse_flags:"))
		{
			if (e_at_SendMouse != pSubAction->e_ActionType)
			{
				String sError;
				sError << "[line " << iLine << "] mouse_flags option found in non mouse action.";
				Error::AddError (sError);
			}
			rLine.Replace ("mouse_flags:", "");
			rLine.Strip ();
			pSubAction->s_MouseFlags = rLine;
		}
		else if (true == rLine.Has ("key_flags:"))
		{
			if (e_at_SendKey != pSubAction->e_ActionType)
			{
				String sError;
				sError << "[line " << iLine << "] key_flags option found in non keyboard action.";
				Error::AddError (sError);
			}
			rLine.Replace ("key_flags:", "");
			rLine.Strip ();
			pSubAction->s_KeyFlags = rLine;
		}
		else if (true == rLine.Has ("virtual_key:"))
		{
			if (e_at_SendKey != pSubAction->e_ActionType)
			{
				String sError;
				sError << "[line " << iLine << "] virtual_key option found in non keyboard action.";
				Error::AddError (sError);
			}
			rLine.Replace ("virtual_key:", "");
			rLine.Strip ();
			pSubAction->s_VirtualKey = rLine;
		}
		else if (true == rLine.Has ("scan_code:"))
		{
			String sError;
			sError << "[line " << iLine << "] scan_code option not supported.";
			Error::AddError (sError);
		}
		else
		{
			pAction = new Action;
			map_NameToAction.insert (make_pair (rLine, pAction));
			continue;
		}
	}
	return true;
}


//
void Action::Clear (void)
{
	Action::map_NameToMessage.clear ();
	Action::map_NameToVirtualKey.clear ();
	ITERATE (NameToAction_map_t, map_NameToAction, ite)
		delete ite->second;
	Action::map_NameToAction.clear ();
}

