#include "stdafx.h"
#include "Identify.h"
#include "IdentifyDoc.h"
#include "UIObject.h"
#include "ConnectionSettingsDlg.h"
#include "InteractDlg.h"
#include "Action.h"
#include "Error.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif


//
IMPLEMENT_DYNCREATE(IdentifyDoc, CDocument)

BEGIN_MESSAGE_MAP(IdentifyDoc, CDocument)
	ON_COMMAND(IDM_SCAN_SYSTEM, &IdentifyDoc::OnScanSystem)
	ON_COMMAND(ID_FILE_CONNECTTOSERVER, &IdentifyDoc::OnFileConnectToServer)
	ON_COMMAND(IDM_FIND_UI_OBJECT, &IdentifyDoc::OnFindUiObject)
	ON_COMMAND(ID_CLIENT_ENABLEDEBUGMODE, &IdentifyDoc::OnClientEnabledebugmode)
	ON_UPDATE_COMMAND_UI(ID_CLIENT_ENABLEDEBUGMODE, &IdentifyDoc::OnUpdateClientEnabledebugmode)
	ON_COMMAND(ID_CLIENT_DISPLAYPROGRESS, &IdentifyDoc::OnClientDisplayprogress)
	ON_UPDATE_COMMAND_UI(ID_CLIENT_DISPLAYPROGRESS, &IdentifyDoc::OnUpdateClientDisplayprogress)
END_MESSAGE_MAP()


IdentifyDoc* IdentifyDoc::p_Document = NULL;


//
IdentifyDoc::IdentifyDoc()
{
	p_PreviousConfiguration = NULL;
	p_CurrentConfiguration = NULL;
	p_IdleStateDiffStart = NULL;
	p_StartingConfiguration = NULL;
	p_Socket = NULL;
	b_EnableHumanReadableMode = false;
	b_DisplayProgress = false;
	b_GotDisconnected = false;
	b_DisconnectedOnRequest = false;
}

IdentifyDoc::~IdentifyDoc()
{
	if (NULL != p_PreviousConfiguration)
		delete p_PreviousConfiguration;
	p_PreviousConfiguration = NULL;
	if (NULL != p_CurrentConfiguration)
		delete p_CurrentConfiguration;
	p_CurrentConfiguration = NULL;
	if (NULL != p_IdleStateDiffStart)
		delete p_IdleStateDiffStart;
	p_IdleStateDiffStart = NULL;
	if (NULL != p_StartingConfiguration)
		delete p_StartingConfiguration;
	p_StartingConfiguration = NULL;

	if (NULL != p_Socket)
		delete p_Socket;
	p_Socket = NULL;

	Request::Clear ();
}


BOOL IdentifyDoc::OnNewDocument()
{
	if (!CDocument::OnNewDocument())
		return FALSE;

	return TRUE;
}


//
bool IdentifyDoc::LoadConfigs (void)
{
	String_dq_t	dqLines;
	if (false == File::ReadLines ("server.cfg", dqLines))
	{
		AfxMessageBox ("Failed to load configuration file 'server.cfg'. Unable to continue.");
		return false;
	}
	ITERATE (String_dq_t, dqLines, ite)
	{
		if (false == (*ite).Has ("="))
			continue;
		String_dq_t	dqTokens;
		(*ite).Split (dqTokens, '=');
		dqTokens [1].Strip ();

		if (true == (*ite).Has ("server_ip"))
			s_ServerIP = dqTokens [1];
		else if (true == (*ite).Has ("server_port"))
			i_ServerPort = dqTokens [1];
	}
	return true;
}


//
void IdentifyDoc::Initialize (void)
{
	if (NULL != IdentifyDoc::p_Document)
		return;

	CWnd* pMainWnd = AfxGetMainWnd ();
	DWORD dwProcessID = 0;
	GetWindowThreadProcessId (pMainWnd->GetSafeHwnd (), &dwProcessID);
	UIObject::dw_MyProcessID = dwProcessID;
	UIObject::LoadIgnoreList ();
	
	Action::LoadConfigs ();
	IdentifyDoc::p_Document = this;

	if (false == IdentifyDoc::LoadConfigs ())
		return;
	UpdateAllViews (NULL, UPDATE_REQUEST_VIEW);

	SetTimer (NULL, 0, 200, &IdentifyDoc::StartupTimerProc);
}


//
VOID CALLBACK IdentifyDoc::StartupTimerProc(HWND _hWnd, UINT _uMsg, UINT_PTR _idEvent, DWORD _dwTime)
{
	KillTimer (NULL, _idEvent);

	CWnd* pMainWnd = AfxGetMainWnd ();
	pMainWnd->ShowWindow (SW_MINIMIZE);
	p_Document->ScanSystem (false);
	p_Document->o_SystemMonitor.Init ();
	p_Document->b_IdleState = true;

	p_Document->SetStartPoint ();

	if (NULL == p_Document->p_Socket)
		p_Document->p_Socket = new ConnectionSocket (p_Document, NULL);
	if (false == p_Document->p_Socket->IsConnected ())
		p_Document->p_Socket->Connect (p_Document->s_ServerIP, p_Document->i_ServerPort);

	SetTimer (NULL, 0, 100, &IdentifyDoc::SystemMonitorTimerProc);
}


//
VOID CALLBACK IdentifyDoc::SystemMonitorTimerProc (HWND _hWnd, UINT _uMsg, UINT_PTR _idEvent, DWORD _dwTime)
{
	// check & reconnect to server if necessary ...
	if (false == p_Document->b_DisconnectedOnRequest)
	{
		if (NULL == p_Document->p_Socket)
			p_Document->p_Socket = new ConnectionSocket (p_Document, NULL);
		if (false == p_Document->p_Socket->IsConnected ())
			p_Document->p_Socket->Connect (p_Document->s_ServerIP, p_Document->i_ServerPort);
	}

	// monitor state ...
	if (true == p_Document->b_IdleState)
		p_Document->o_SystemMonitor.MeasureAverageLoad ();

	else
	{
		// check if the system has settled down ...
		if (0 <= p_Document->o_SystemMonitor.CheckCurrentLoad ())
		{
			//TRACE ("   load too high ...\n");
			return;
		}

		if (NULL == p_Document->p_IdleStateDiffStart)
		{
			// TRACE ("   first window snapshot ...\n");
			p_Document->b_WindowsStatic = false;
			p_Document->p_IdleStateDiffStart = new SystemConfiguration;
			CWnd* pDesktopWnd = CWnd::GetDesktopWindow ();

			UIObject::p_CurrentSystemConfig = p_Document->p_IdleStateDiffStart;
			UIObject::ScanFromWindow (pDesktopWnd, &p_Document->set_ProcessesToIgnore);
			p_Document->p_IdleStateDiffStart->CreateWindowTree ();
			return;
		}
		if (false == p_Document->b_WindowsStatic)
		{
			// check if the windows have stopped changing ...
			SystemConfiguration* pConfiguration = new SystemConfiguration;
			CWnd* pDesktopWnd = CWnd::GetDesktopWindow ();

			UIObject::p_CurrentSystemConfig = pConfiguration;
			UIObject::ScanFromWindow (pDesktopWnd, &p_Document->set_ProcessesToIgnore);
			pConfiguration->CreateWindowTree ();
			if (true == pConfiguration->IsChanging (*p_Document->p_IdleStateDiffStart, p_Document->set_ProcessesToIgnore))
			{
				//TRACE ("   windows still changing ...\n");
				delete p_Document->p_IdleStateDiffStart;
				p_Document->p_IdleStateDiffStart = pConfiguration;
				return;
			}
			delete pConfiguration;
			p_Document->b_WindowsStatic = true;
		}

		//TRACE ("   returned to idle.\n");
		delete p_Document->p_IdleStateDiffStart;
		p_Document->p_IdleStateDiffStart = NULL;

		// reset to idle state ...
		p_Document->b_IdleState = true;
		p_Document->o_SystemMonitor.MeasureAverageLoad ();
		bool bEnvironmentChanged = p_Document->ScanSystem (false);

		Sleep (500);
		if (NULL != p_Document->p_Socket)
		{
			String sText;
			if (true == p_Document->b_EnableHumanReadableMode)
			{
				sText << "   action completed";
				if (true == bEnvironmentChanged)
					sText << ", env has changed";
				sText << ".\n--------------------------------\x10\n";
			}
			else
			{
				sText << "   action completed";
				if (true == bEnvironmentChanged)
					sText << ", env has changed";
				sText << ".\x10\n";
			}
			p_Document->p_Socket->Send (sText);
		}
	}
}


//
bool IdentifyDoc::ScanSystem (bool _bPreserveLastDiff)
{
	// get desktop window
	CWnd* pDesktopWnd = CWnd::GetDesktopWindow ();
	if (NULL == pDesktopWnd)
	{
		AfxMessageBox ("Could not retrieve the desktop window! Scan aborted.", MB_ICONSTOP|MB_OK);
		return false;
	}

	if (false == _bPreserveLastDiff)
	{
		if (NULL != p_PreviousConfiguration)
			delete p_PreviousConfiguration;
		p_PreviousConfiguration = p_CurrentConfiguration;
	}
	else
	{
		if (NULL != p_CurrentConfiguration)
			delete p_CurrentConfiguration;
	}
	p_CurrentConfiguration = new SystemConfiguration;

	UIObject::p_CurrentSystemConfig = p_CurrentConfiguration;
	UIObject::ScanFromWindow (pDesktopWnd);
	p_CurrentConfiguration->CreateWindowTree ();

	bool bDifferent = false;
	if (NULL != p_PreviousConfiguration)
		bDifferent = p_CurrentConfiguration->Diff (*p_PreviousConfiguration, _bPreserveLastDiff);

	if (true == b_DisplayProgress)
		UpdateAllViews (NULL, UPDATE_WINDOW_VIEWS);
	return bDifferent;
}


//
void IdentifyDoc::OnScanSystem (void)
{
	// Error::sdq_Errors.clear ();
	ScanSystem (false);
	UpdateAllViews (NULL, UPDATE_REQUEST_VIEW);
}


//
void IdentifyDoc::OnFileConnectToServer()
{
	ConnectionSettingsDlg dlg;
	dlg.s_ServerIP = "172.16.241.1";
	dlg.i_ServerPort = 5000;
	if (IDOK != dlg.DoModal ())
		return;

	s_ServerIP = dlg.s_ServerIP;
	i_ServerPort = dlg.i_ServerPort;
	if (NULL != p_Socket)
		delete p_Socket;
	p_Socket = NULL;
	// p_Socket = new ConnectionSocket (this, NULL);
	// p_Socket->Connect (dlg.s_ServerIP, dlg.i_ServerPort);
}


//
void IdentifyDoc::OnFindUiObject()
{
	InteractDlg	dlg;
	if (IDOK != dlg.DoModal ())
		return;

	ScanSystem (true);
	UIObject_dq_t dqUIObjects;
	String sObjectName ((LPCSTR)dlg.s_WindowName);
	sObjectName.LowerCase ();
	p_CurrentConfiguration->FindApproximate (sObjectName, &dqUIObjects);

	Request* pRequest = new Request;
	pRequest->c_Type = '?';
	pRequest->s_RawText << "?" << sObjectName;
	pRequest->s_UIObjectID = sObjectName;
	UIObject_dq_t::iterator	ite;
	for (ite = dqUIObjects.begin ();
		ite != dqUIObjects.end ();
		++ ite)
	{
		pRequest->dq_AffectedUIObjectIDs.push_back ((*ite)->s_ID);
	}

	if (true == b_DisplayProgress)
		UpdateAllViews (NULL, UPDATE_REQUEST_VIEW);
}


//
void IdentifyDoc::PerformAction (const char* _zObjectID, const char* _zAction, const char* _zText, Request* _pRequest)
{
	UIObject* pUIObject = NULL;
	if (0 != strcmp ("*", _zObjectID))
		pUIObject = p_CurrentConfiguration->Find (_zObjectID);

	String sAction (_zAction);
	if (false == sAction.Has ("type"))
		b_IdleState = false;

	// After performing the requested action, we need to check if the windows have stopped changing.
	// Here we can ignore all old windows that are not associated with the process we are going to 
	// interact with ...
	set_ProcessesToIgnore = UIObject::set_ProcessIDs;
	if (NULL != pUIObject)
		set_ProcessesToIgnore.erase (pUIObject->dw_ProcessID);
	else if (NULL != p_CurrentConfiguration->p_ForegroundObject)
		set_ProcessesToIgnore.erase (p_CurrentConfiguration->p_ForegroundObject->dw_ProcessID);

	//
	if ((NULL != pUIObject) && (UIOBJECT__COMBOBOX_ITEM == pUIObject->c_ObjectType))
	{
		if ((NULL != pUIObject->p_ActivationWindow) && (false == pUIObject->p_ActivationWindow->b_IsForeground))
		{
			CWnd* pActivationWnd = CWnd::FromHandle (pUIObject->p_ActivationWindow->h_Wnd);
			pActivationWnd->SetForegroundWindow ();
		}
		CWnd* pWnd = CWnd::FromHandle (pUIObject->h_ParentWnd);
		pWnd->SendMessage (CB_SETCURSEL, pUIObject->i_CommandID, 0);
		//int ret = pWnd->SendMessage (CB_SELECTSTRING, -1, pUIObject->s_Name);
		UIObject* pCombobox = pUIObject->p_Parent;
		//SendMessage (pCombobox->p_Parent->h_Wnd, WM_NOTIFY, (WPARAM)pCombobox->i_CommandID, NULL);
		TRACE ("\n");
	}
	else
	{
		//if ((NULL != pUIObject->p_ActivationWindow) && (false == pUIObject->p_ActivationWindow->b_IsForeground))
		//{
		//	CWnd* pActivationWnd = CWnd::FromHandle (pUIObject->p_ActivationWindow->h_Wnd);
		//	pActivationWnd->SetForegroundWindow ();
		//}
		if ("select" == sAction)
			return;


		CWnd* pWnd = NULL;
		CPoint pos;
		if (NULL != pUIObject)
		{
			if ((UIOBJECT__TOOLBAR_BUTTON == pUIObject->c_ObjectType) ||
				(UIOBJECT__TABCTRL_BUTTON == pUIObject->c_ObjectType))
			{
				pWnd = CWnd::FromHandle (pUIObject->h_ParentWnd);
				pos = pUIObject->rct_Body.CenterPoint ();
				pWnd->ClientToScreen (&pos);
			}
			else if (UIOBJECT__LISTVIEW_BUTTON == pUIObject->c_ObjectType)
			{
				pWnd = CWnd::FromHandle (pUIObject->h_ParentWnd);
				
				// make the list view item visible...
				pWnd->SendMessage (LVM_ENSUREVISIBLE, pUIObject->i_CommandID, 0);

				// get it's rectangle ...
				DWORD iProcessID = 0;
				GetWindowThreadProcessId (pUIObject->h_ParentWnd, &iProcessID);
				HANDLE hProcess = OpenProcess (PROCESS_VM_READ|PROCESS_VM_WRITE|PROCESS_VM_OPERATION, 0, iProcessID);
				ASSERT (NULL != hProcess);

				// allocated memory accessible by the process we retrieved above ...
				LPVOID lpRemoteRect = VirtualAllocEx (hProcess, NULL, sizeof (RECT), MEM_COMMIT, PAGE_READWRITE);
				if (TRUE == SendMessage (pUIObject->h_ParentWnd, LVM_GETITEMRECT, (WPARAM)pUIObject->i_CommandID, (LPARAM)lpRemoteRect))
				{
					RECT rect;
					DWORD iBytesRead;
					ReadProcessMemory (hProcess, lpRemoteRect, &rect, sizeof(RECT), &iBytesRead);
					pUIObject->rct_Body = rect;
				}
				VirtualFreeEx (hProcess, lpRemoteRect, 0, MEM_RELEASE);
				//

				pos = pUIObject->rct_Body.CenterPoint ();
				if ((LVS_REPORT == (LVS_TYPEMASK & pUIObject->p_Parent->dw_Style)) ||
					(LVS_LIST == (LVS_TYPEMASK & pUIObject->p_Parent->dw_Style)))
					pos.x = (pUIObject->rct_Body.left + 40 < pos.x)? pUIObject->rct_Body.left + 40 : pos.x;
				pWnd->ClientToScreen (&pos);
			}
			else if (UIOBJECT__TREEVIEW_ITEM == pUIObject->c_ObjectType)
			{
				pWnd = CWnd::FromHandle (pUIObject->h_ParentWnd);
				
				// make the list view item visible...
				pWnd->SendMessage (TVM_ENSUREVISIBLE, 0, (LPARAM)pUIObject->i_CommandID);

				// get it's rectangle ...
				DWORD iProcessID = 0;
				GetWindowThreadProcessId (pUIObject->h_ParentWnd, &iProcessID);
				HANDLE hProcess = OpenProcess (PROCESS_VM_READ|PROCESS_VM_WRITE|PROCESS_VM_OPERATION, 0, iProcessID);
				ASSERT (NULL != hProcess);

				// allocated memory accessible by the process we retrieved above ...
				LPVOID lpRemoteRect = VirtualAllocEx (hProcess, NULL, sizeof (RECT), MEM_COMMIT, PAGE_READWRITE);
				DWORD iBytesWritten;
				WriteProcessMemory (hProcess, lpRemoteRect, &(pUIObject->i_CommandID), sizeof (long), &iBytesWritten);
				if (TRUE == SendMessage (pUIObject->h_ParentWnd, TVM_GETITEMRECT, (WPARAM)TRUE, (LPARAM)lpRemoteRect))
				{
					RECT rect;
					DWORD iBytesRead;
					ReadProcessMemory (hProcess, lpRemoteRect, &rect, sizeof(RECT), &iBytesRead);
					pUIObject->rct_Body = rect;
				}
				VirtualFreeEx (hProcess, lpRemoteRect, 0, MEM_RELEASE);
				//

				pos = pUIObject->rct_Body.CenterPoint ();
				pWnd->ClientToScreen (&pos);
			}
			else if (UIOBJECT__MENU_ITEM == pUIObject->c_ObjectType)
			{
				pWnd = CWnd::FromHandle (pUIObject->h_ParentWnd);
				pos = pUIObject->rct_Body.CenterPoint ();
			}
			else
			{
				pWnd = CWnd::FromHandle (pUIObject->h_Wnd);
				pos = pUIObject->rct_Body.CenterPoint ();
			}
		}

		CWnd* pDesktop = CWnd::GetDesktopWindow ();
		CRect rectDesktop;
		pDesktop->GetWindowRect (&rectDesktop);
		pos.x = (LONG) (pos.x * 65535 / (double) rectDesktop.Width ());
		pos.y = (LONG) (pos.y * 65535 / (double) rectDesktop.Height ());

		Action* pAction = Action::Find (sAction);
		if (NULL == pAction)
			return;
		pAction->Act (pUIObject, pWnd, pos, _zText);
	}

	if (true == sAction.Has ("type"))
	{
		String sText = "   action completed.\x10\n";
		p_Socket->Send (sText);
	}
	else
		Sleep (100);
}


//
void IdentifyDoc::ReplyWithObjectHumanReadable (const char* _zSearchName, UIObject* _pObject)
{
	String sText;
	sText << "[ " << _pObject->s_ID << " ][ " << _pObject->p_MainWindow->s_ID << " ][ " << _pObject->p_ActivationWindow->s_ID << " ] "
		  << "nme:(\x01" << _pObject->s_Name << "\x01)"
		  << '~' << ((true == _pObject->o_Diff.bf_Flags.b_Name)? 1 : 0) << "  "
		  << "cls:(\x01" << _pObject->s_ClassName << "\x01)  "
		  << "edd:" << String::LevenshteinDistance (_pObject->s_LowercaseName, _zSearchName) << "\n";
	if (NULL != _pObject->p_MainWindow)
	{
		sText << "\tmwd:(\x01" << _pObject->p_MainWindow->s_Name << "\x01)"
			  << '~' << ((true == _pObject->p_MainWindow->o_Diff.bf_Flags.b_Name)? 1 : 0) << "  "
			  << "mss:" << _pObject->p_MainWindow->GetShowState () 
			  << '~' << ((true == _pObject->p_MainWindow->o_Diff.bf_Flags.b_ShowState)? 1 : 0) << "  "
			  << "~mwd-new:" << ((true == _pObject->p_MainWindow->o_Diff.bf_Flags.b_New)? 1 : 0) << "  "
			  << "mwd-edd:" << String::LevenshteinDistance (_pObject->p_MainWindow->s_LowercaseName, _zSearchName) << "\n";
	}
	if (NULL != _pObject->p_ActivationWindow)
	{
		sText << "\tawd:(\x01" << _pObject->p_ActivationWindow->s_Name << "\x01)"
			  << '~' << ((true == _pObject->p_ActivationWindow->o_Diff.bf_Flags.b_Name)? 1 : 0) << "  "
			  << "ass:" << _pObject->p_ActivationWindow->GetShowState () 
			  << '~' << ((true == _pObject->p_ActivationWindow->o_Diff.bf_Flags.b_ShowState)? 1 : 0) << "  "
			  << "~awd-new:" << ((true == _pObject->p_ActivationWindow->o_Diff.bf_Flags.b_New)? 1 : 0) << "  "
			  << "awd-edd:" << String::LevenshteinDistance (_pObject->p_ActivationWindow->s_LowercaseName, _zSearchName) << "\n";
	}
	sText << "\tvsb:" << ((true == _pObject->b_IsVisible)? 1 : 0) << '~'
		  << ((true == _pObject->o_Diff.bf_Flags.b_IsVisible)? 1 : 0) << "  "
		  << "fcs:" << ((true == _pObject->b_HasFocus)? 1 : 0) << '~'
		  << ((true == _pObject->o_Diff.bf_Flags.b_HasFocus)? 1 : 0) << "  "
		  << "act:" << ((true == _pObject->b_IsActive)? 1 : 0) << '~'
		  << ((true == _pObject->o_Diff.bf_Flags.b_IsActive)? 1 : 0) << "  "
		  << "fgr:" << ((true == _pObject->b_IsForeground)? 1 : 0) << '~'
		  << ((true == _pObject->o_Diff.bf_Flags.b_IsForeground)? 1 : 0) << "  "
		  << "ien:" << ((true == _pObject->b_IsEnabledForInput)? 1 : 0) << '~'
		  << ((true == _pObject->o_Diff.bf_Flags.b_IsEnabledForInput)? 1 : 0) << "  "
		  << "cld:" << ((true == _pObject->b_IsChildWnd)? 1 : 0) << "  "
		  << "sst:" << _pObject->GetShowState () << '~'
		  << ((true == _pObject->o_Diff.bf_Flags.b_ShowState)? 1 : 0) << "  "
		  << "sty:" << (long) _pObject->dw_Style << '~'
		  << ((true == _pObject->o_Diff.bf_Flags.b_Style)? 1 : 0) << "  "
		  << "exs:" << (long) _pObject->dw_ExStyle << '~'
		  << ((true == _pObject->o_Diff.bf_Flags.b_ExStyle)? 1 : 0) << '\n'
		  << "\tmenu:" << (long)_pObject->i_MenuItemState << "  "
		  << "~menu:" << ((true == _pObject->o_Diff.bf_Flags.b_MenuItemState)? 1 : 0) << "  "
		  << "~rect:" << ((true == _pObject->o_Diff.bf_Flags.b_RectBody)? 1 : 0) << "  "
		  << "~new:" << ((true == _pObject->o_Diff.bf_Flags.b_New)? 1 : 0) << "  "
		  << "~missing:" << ((true == _pObject->o_Diff.bf_Flags.b_Missing)? 1 : 0) << "  "
		  << "~child:" << ((true == _pObject->o_Diff.bf_Flags.b_ChildIsDifferent)? 1 : 0) << "\x02\n";

	p_Socket->Send (sText);
}


//
void IdentifyDoc::ReplyWithObject (const char* _zSearchName, UIObject* _pObject)
{
	String sName = _pObject->s_Name;
	if ("" != _pObject->s_ExtendedName)
		sName = _pObject->s_ExtendedName;
	double dNormalizedEditDist = -1;
	if (NULL != _zSearchName)
	{
		long lEditDist = String::LevenshteinDistance (_pObject->s_LowercaseName, _zSearchName);
		dNormalizedEditDist = lEditDist / (double)(_pObject->s_LowercaseName.length () + strlen (_zSearchName));

		if (("" != _pObject->s_ExtendedName) && (_pObject->s_LowercaseName != _pObject->s_ExtendedName))
		{
			long lAltEditDist = String::LevenshteinDistance (_pObject->s_ExtendedName, _zSearchName);
			double dAltNormalizedEditDist = lAltEditDist / (double)(_pObject->s_ExtendedName.length () + strlen (_zSearchName));
			if (dNormalizedEditDist > dAltNormalizedEditDist)
				dNormalizedEditDist = dAltNormalizedEditDist;
		}
	}

	int iObjectFamilyPreviouslyTouched = 0;
	int iMainWindowPreviouslyTouched = 0;
	if (set_TouchedObjects.end () != set_TouchedObjects.find (_pObject->p_MainWindow->s_ID))
	{
		iMainWindowPreviouslyTouched = 1;
		iObjectFamilyPreviouslyTouched = 1;
	}
	int iActWindowPreviouslyTouched = 0;
	if (set_TouchedObjects.end () != set_TouchedObjects.find (_pObject->p_ActivationWindow->s_ID))
	{
		iActWindowPreviouslyTouched = 1;
		iObjectFamilyPreviouslyTouched = 2;
	}
	if (NULL != _pObject->p_Parent)
	{
		if (set_TouchedObjects.end () != set_TouchedObjects.find (_pObject->p_Parent->s_ID))
			iObjectFamilyPreviouslyTouched = 3;
	}
	int iObjectPreviouslyTouched = 0;
	if (set_TouchedObjects.end () != set_TouchedObjects.find (_pObject->s_ID))
	{
		iObjectPreviouslyTouched = 1;
		// iObjectFamilyPreviouslyTouched = 4;
	}

	String sText;
	sText << _pObject->s_ID << '\x01' << _pObject->p_MainWindow->s_ID << '\x01' << _pObject->p_ActivationWindow->s_ID << '\x01'
		  << _pObject->s_Name << '\x01'
		  << _pObject->GetStandardizedClassName () << '\x01'
		  << iObjectFamilyPreviouslyTouched << '\x01' 
		  << iObjectPreviouslyTouched << '\x01' 
		  << iMainWindowPreviouslyTouched << '\x01' 
		  << iActWindowPreviouslyTouched << '\x01'
		  << ((true == _pObject->o_Diff.bf_Flags.b_Name)? 1 : 0) << '\x01'
		  << dNormalizedEditDist << '\x01';

	// combination of diff_visible & visible of all three windows
	bool bInvisibleRect = ((-1 == _pObject->rct_Body.left) &&
							(-1 == _pObject->rct_Body.right) &&
							(-1 == _pObject->rct_Body.top) &&
							(-1 == _pObject->rct_Body.bottom));
	if (NULL != _pObject->p_Parent)
	{
		bInvisibleRect |= ((-1 == _pObject->p_Parent->rct_Body.left) &&
								(-1 == _pObject->p_Parent->rct_Body.right) &&
								(-1 == _pObject->p_Parent->rct_Body.top) &&
								(-1 == _pObject->p_Parent->rct_Body.bottom));
	}
	bInvisibleRect |= ((-1 == _pObject->p_ActivationWindow->rct_Body.left) &&
							(-1 == _pObject->p_ActivationWindow->rct_Body.right) &&
							(-1 == _pObject->p_ActivationWindow->rct_Body.top) &&
							(-1 == _pObject->p_ActivationWindow->rct_Body.bottom));

	bool bVisible = ((true == _pObject->b_IsVisible) &&
					(true == _pObject->p_ActivationWindow->b_IsVisible) &&
					(false == bInvisibleRect)); //&&
					//(true == pUIObject->p_MainWindow->b_IsVisible));
	bool bNewlyVisible = ((true == _pObject->o_Diff.bf_Flags.b_IsVisible) ||
						(true == _pObject->p_ActivationWindow->o_Diff.bf_Flags.b_IsVisible)); //||
						//(true == pUIObject->p_MainWindow->o_Diff.bf_Flags.b_IsVisible));
	sText << ((true == bVisible)? 1 : 0) << '\x01'
		  << (((true == bNewlyVisible) && (true == bVisible))? 1 : 0) << '\x01';

	// combination of diff_showstate & showstate of all three windows
	bool bShowing = (("show" == _pObject->GetShowState ()) &&
					("show" == _pObject->p_ActivationWindow->GetShowState ())); //&&
					//("show" == pUIObject->p_MainWindow->GetShowState ()));
	bool bNewlyShowing = ((true == _pObject->o_Diff.bf_Flags.b_ShowState) ||
						(true == _pObject->p_ActivationWindow->o_Diff.bf_Flags.b_ShowState)); //||
						//(true == pUIObject->p_MainWindow->o_Diff.bf_Flags.b_ShowState));
	sText << ((true == bShowing)? 1 : 0) << '\x01'
		  << (((true == bNewlyShowing) && (true == bShowing))? 1 : 0) << '\x01';

	// combination of diff_focus & focus of all three windows
	bool bInFocus = ((true == _pObject->b_HasFocus) ||
					(true == _pObject->p_ActivationWindow->b_HasFocus)); //||
					//(true == pUIObject->p_MainWindow->b_HasFocus));
	bool bNewlyInFocus = ((true == _pObject->o_Diff.bf_Flags.b_HasFocus) ||
						(true == _pObject->p_ActivationWindow->o_Diff.bf_Flags.b_HasFocus)); //||
						//(true == pUIObject->p_MainWindow->o_Diff.bf_Flags.b_HasFocus));
	sText << ((true == bInFocus)? 1 : 0) << '\x01'
		  << (((true == bNewlyInFocus) && (true == bInFocus))? 1 : 0) << '\x01';

	// combination of diff_fg & fg of all three windows
	bool bFg = ((true == _pObject->b_IsForeground) ||
				(true == _pObject->p_ActivationWindow->b_IsForeground));// ||
				//(true == pUIObject->p_MainWindow->b_IsForeground));
	bool bNewlyFg = ((true == _pObject->o_Diff.bf_Flags.b_IsForeground) ||
					(true == _pObject->p_ActivationWindow->o_Diff.bf_Flags.b_IsForeground)); //||
					//(true == pUIObject->p_MainWindow->o_Diff.bf_Flags.b_IsForeground));
	sText << ((true == bFg)? 1 : 0) << '\x01'
		  << (((true == bNewlyFg) && (true == bFg))? 1 : 0) << '\x01';

	// enabled for input & diff
	bool bNewlyEI = ((true == _pObject->o_Diff.bf_Flags.b_IsEnabledForInput) &&
					(true == _pObject->b_IsEnabledForInput));
	sText << ((true == _pObject->b_IsEnabledForInput)? 1 : 0) << '\x01'
		  << ((true == bNewlyEI)? 1 : 0) << '\x01';

	// child
	sText << ((true == _pObject->b_IsChildWnd)? 1 : 0) << '\x01';

	// new
	sText << ((true == _pObject->o_Diff.bf_Flags.b_New)? 1 : 0);

	//
	sText << "\x03\n";
	p_Socket->Send (sText);
}


//
void IdentifyDoc::Reply (const char* _zSearchName, UIObject_dq_t& _rdqUIObjects)
{
	if (NULL == p_Socket)
		return;

	UIObject_dq_t::iterator	ite;
	for (ite = _rdqUIObjects.begin ();
		ite != _rdqUIObjects.end ();
		++ ite)
	{
		UIObject* pUIObject = *ite;
		if (true == pUIObject->o_Diff.bf_Flags.b_Missing)
			continue;

		if (true == b_EnableHumanReadableMode)
			ReplyWithObjectHumanReadable (_zSearchName, pUIObject);
		else
		{
			ReplyWithObject (_zSearchName, pUIObject);
			String_map_t::iterator	iteOldName;
			iteOldName = UIObject::map_IdToOldName.find (pUIObject->s_ID);
			if (iteOldName != UIObject::map_IdToOldName.end ())
			{
				String sOriginalName = pUIObject->s_Name;
				pUIObject->s_Name = iteOldName->second;
				ReplyWithObject (_zSearchName, pUIObject);
				pUIObject->s_Name = sOriginalName;
			}
		}
	}
	String sTerminal;
	if (true == b_EnableHumanReadableMode)
		sTerminal = "--------------------------------\x10\n";
	else
		sTerminal = "\x10\n";
	p_Socket->Send (sTerminal);
}



//
void IdentifyDoc::OnConnect (void)
{
	Error::AddError ("socket connected");
	if (true == b_GotDisconnected)
	{
		b_GotDisconnected = false;
		p_Socket->Send ("\n");
		return;
	}
	b_GotDisconnected = false;

	Error::AddError ("sending [restart complete]");
	String sText;
	if (true == b_EnableHumanReadableMode)
		sText = "retart complete.\n--------------------------------\x10\n";
	else
		sText = "restart complete.\x10\n";
	p_Socket->Send (sText);
	UpdateAllViews (NULL, UPDATE_REQUEST_VIEW);
}



//
void IdentifyDoc::OnDisconnect (void)
{
	Error::AddError ("socket disconnected");
	b_GotDisconnected = true;
	p_Socket = NULL;
}


//
#define REQUEST__FIND_UI_OBJECT		(char)'?'
#define REQUEST__LIST_UI_OBJECT		(char)'*'
#define REQUEST__PERFORM_ACTION		(char)'!'
#define REQUEST__SET_START			(char)'>'
#define REQUEST__RESET_TO_START		(char)'<'
#define REQUEST__DISCONNECT			(char)'@'


//
#define MRK_ACTION		':'
#define MRK_TEXT		'|'
#define MRK_TERM		'\n'

//
void IdentifyDoc::OnReceive (const char* _zBuffer, int _iBytes)
{
	char* pRequestText = NULL;
	s_ReceiveBuffer << _zBuffer;
	s_ReceiveBuffer.Replace ("\n", "");

	if (NULL == strchr (_zBuffer, MRK_TERM))
		// if the message terminator is not present, remember the data received thus far ...
		return;
	else
	{
		s_ReceiveBuffer.Strip ();
		pRequestText = (char*)(const char*)s_ReceiveBuffer;
	}

	char* pTerm = (char*) strchr (_zBuffer, MRK_TERM);
	*pTerm = '\0';

	//
	if (REQUEST__SET_START == *pRequestText)
	{
		if (true == b_DisplayProgress)
		{
			Request* pRequest = new Request;
			pRequest->c_Type = '>';
			pRequest->s_RawText = pRequestText;
			pRequest->s_UIObjectID = "set starting point";
		}
		SetStartPoint ();
	}
	else if (REQUEST__RESET_TO_START == *pRequestText)
	{
		if (true == b_DisplayProgress)
		{
			Request* pRequest = new Request;
			pRequest->c_Type = '<';
			pRequest->s_RawText = pRequestText;
			pRequest->s_UIObjectID = "reset";
		}
		ResetToStartPoint ();
		set_TouchedObjects.clear ();
	}
	else if (REQUEST__LIST_UI_OBJECT == *pRequestText)
	{
		ScanSystem (true);
		UIObject_dq_t dqUIObjects;
		p_CurrentConfiguration->ListAllObjects (&dqUIObjects);
		Reply (NULL, dqUIObjects);
		UIObject::ClearTemporaryObjects ();
	}
	else if (REQUEST__FIND_UI_OBJECT == *pRequestText)
	{
		Request* pRequest = NULL;
		if (true == b_DisplayProgress)
		{
			pRequest = new Request;
			pRequest->c_Type = '?';
			pRequest->s_RawText = pRequestText;
		}
		char* pObjectName = pRequestText;
		while ((' ' == *pObjectName) || ('\t' == *pObjectName))
			++ pObjectName;
		if (NULL == pObjectName)
		{
			// TRACE ("[ERROR]  malformed object query! '%s'", pRequestText);
			return;
		}
		++ pObjectName;

		ScanSystem (true);
		UIObject_dq_t dqUIObjects;
		String sObjectName (pObjectName);
		sObjectName.LowerCase ();
		p_CurrentConfiguration->FindApproximate (sObjectName, &dqUIObjects);
		Reply (sObjectName, dqUIObjects);

		if (true == b_DisplayProgress)
		{
			pRequest->s_UIObjectID = pObjectName;
			UIObject_dq_t::iterator	ite;
			for (ite = dqUIObjects.begin ();
				ite != dqUIObjects.end ();
				++ ite)
			{
				pRequest->dq_AffectedUIObjectIDs.push_back ((*ite)->s_ID);
			}
		}
		UIObject::ClearTemporaryObjects ();
	}

	//
	else if (REQUEST__PERFORM_ACTION == *pRequestText)
	{
		Request* pRequest = NULL;
		if (true == b_DisplayProgress)
		{
			pRequest = new Request;
			pRequest->c_Type = '!';
			pRequest->s_RawText = pRequestText;
		}
		char* pObjectID = pRequestText;
		while ((' ' == *pObjectID) || ('\t' == *pObjectID))
			++ pObjectID;

		if (NULL == pObjectID)
		{
			if (true == b_DisplayProgress)
				pRequest->s_Error = "[ERROR]  malformed action request.";
			p_Socket->Send ("[ERROR]  malformed action request.\x10\n");
			return;
		}
		++ pObjectID;

		char* pActionType = strchr (pObjectID, MRK_ACTION);
		if (NULL == pActionType)
		{
			if (true == b_DisplayProgress)
				pRequest->s_Error = "[ERROR]  malformed action request.";
			p_Socket->Send ("[ERROR]  malformed action request.\x10\n");
			return;
		}
		*pActionType = '\0';
		++ pActionType;

		char* pText = strchr (pActionType, MRK_TEXT);
		if (NULL == pText)
		{
			if (0 == strcmp (pActionType, "set_text"))
			{
				if (true == b_DisplayProgress)
					pRequest->s_Error = "[ERROR]  malformed action request.";
				p_Socket->Send ("[ERROR]  malformed action request.\x10\n");
				return;
			}
		}
		else
		{
			*pText = '\0';
			++ pText;
		}

		UIObject* pUIObject = NULL;
		if (0 != strcmp ("*", pObjectID))
			pUIObject = p_CurrentConfiguration->Find (pObjectID);
		if (true == b_DisplayProgress)
		{
			pRequest->s_UIObjectID = pObjectID;
			pRequest->s_Action = pActionType;
			if (NULL != pText)
				pRequest->s_Data = pText;
			if (NULL != pUIObject)
				pRequest->dq_AffectedUIObjectIDs.push_back (pUIObject->s_ID);
			else
				pRequest->s_Error = "Object ID not found.";
		}
		if (NULL != Action::Find (pActionType))
		{
			PerformAction (pObjectID, pActionType, pText, pRequest);
			if (NULL != pUIObject)
			{
				set_TouchedObjects.insert (pUIObject->s_ID);
				set_TouchedObjects.insert (pUIObject->p_MainWindow->s_ID);
				set_TouchedObjects.insert (pUIObject->p_ActivationWindow->s_ID);
				if (NULL != pUIObject->p_Parent)
					set_TouchedObjects.insert (pUIObject->p_Parent->s_ID);
			}
		}
		else
			p_Socket->Send ("[ERROR]  unknown action type.\x10\n");
	}

	// clear receive buffer ...
	s_ReceiveBuffer = "";

	if (true == b_DisplayProgress)
		UpdateAllViews (NULL, UPDATE_REQUEST_VIEW);
}


//
void IdentifyDoc::OnDestroySocket (void)
{
	p_Socket = NULL;
}


//
void IdentifyDoc::SetStartPoint (void)
{
	CWnd* pDesktopWnd = CWnd::GetDesktopWindow ();
	if (NULL == pDesktopWnd)
		return;

	if (NULL != p_StartingConfiguration)
		delete p_StartingConfiguration;
	p_StartingConfiguration = new SystemConfiguration;

	UIObject::p_CurrentSystemConfig = p_StartingConfiguration;
	UIObject::ScanFromWindow (pDesktopWnd);
	p_StartingConfiguration->CreateWindowTree ();

	String sText;
	if (true == b_EnableHumanReadableMode)
		sText = "starting point set.\n--------------------------------\x10\n";
	else
		sText = "starting point set.\x10\n";
	if (NULL != p_Socket)
		p_Socket->Send (sText);
}


//
void IdentifyDoc::ResetToStartPoint (void)
{
	if (NULL == p_StartingConfiguration)
	{
		String sText;
		if (true == b_EnableHumanReadableMode)
			sText = "cannot reset, no starting point set.\n--------------------------------\x10\n";
		else
			sText = "cannot reset, no starting point set.\x10\n";
		p_Socket->Send (sText);
		return;
	}

	CWnd* pDesktopWnd = CWnd::GetDesktopWindow ();
	if (NULL == pDesktopWnd)
		return;

	SystemConfiguration* pConfiguration;
	String sError;

	// first close the start menu if it is open ...
	pConfiguration = new SystemConfiguration;
	UIObject::p_CurrentSystemConfig = pConfiguration;
	UIObject::ScanFromWindow (pDesktopWnd);
	pConfiguration->CreateWindowTree ();
	pConfiguration->Diff (*p_StartingConfiguration, false);
	pConfiguration->RollbackStartMenu (*p_StartingConfiguration, sError);
	delete pConfiguration;

	// reset all other windows ...
	pConfiguration = new SystemConfiguration;
	UIObject::p_CurrentSystemConfig = pConfiguration;
	UIObject::ScanFromWindow (pDesktopWnd);
	pConfiguration->CreateWindowTree ();
	pConfiguration->Diff (*p_StartingConfiguration, false);

	if (false == pConfiguration->Rollback (*p_StartingConfiguration, sError))
	{
		String sText;
		if (true == b_EnableHumanReadableMode)
			sText << "reset failed. " << sError << "\n--------------------------------\x10\n";
		else
			sText << "reset failed. " << sError << "\x10\n";
		p_Socket->Send (sText);
	}
	else
	{
		String sText;
		if (true == b_EnableHumanReadableMode)
			sText = "reset complete.\n--------------------------------\x10\n";
		else
			sText = "reset complete.\x10\n";
		p_Socket->Send (sText);
	}
	delete pConfiguration;
}


//
void IdentifyDoc::OnClientEnabledebugmode()
{
	if (true == b_EnableHumanReadableMode)
		b_EnableHumanReadableMode = false;
	else
		b_EnableHumanReadableMode = true;
}
void IdentifyDoc::OnUpdateClientEnabledebugmode(CCmdUI *pCmdUI)
{
	if (true == b_EnableHumanReadableMode)
		pCmdUI->SetCheck (TRUE);
	else
		pCmdUI->SetCheck (FALSE);
}

//
void IdentifyDoc::OnClientDisplayprogress()
{
	if (true == b_DisplayProgress)
		b_DisplayProgress = false;
	else
		b_DisplayProgress = true;
}
void IdentifyDoc::OnUpdateClientDisplayprogress(CCmdUI *pCmdUI)
{
	if (true == b_DisplayProgress)
		pCmdUI->SetCheck (TRUE);
	else
		pCmdUI->SetCheck (FALSE);
}



//
void IdentifyDoc::Serialize(CArchive& ar)
{
}


//
#ifdef _DEBUG
void IdentifyDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void IdentifyDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif

