#include "stdafx.h"
#include "Identify.h"

#include "IdentifyDoc.h"
#include "IdentifyView.h"
#include "InteractDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif


//
IMPLEMENT_DYNCREATE(IdentifyView, CTreeView)
BEGIN_MESSAGE_MAP(IdentifyView, CTreeView)
	ON_NOTIFY_REFLECT(NM_RCLICK, &IdentifyView::OnNMRClick)
	ON_COMMAND(IDM_INTERACT_WITH_WINDOW, &IdentifyView::OnInteractWithWindow)
END_MESSAGE_MAP()


//
IdentifyView::IdentifyView()
{
	h_SelectedItem = NULL;
}

IdentifyView::~IdentifyView()
{
}


//
void IdentifyView::OnInitialUpdate (void)
{
	CTreeCtrl& rTreeCtrl = CTreeView::GetTreeCtrl ();
	rTreeCtrl.ModifyStyle (0, TVS_HASLINES|TVS_LINESATROOT|TVS_HASBUTTONS);

	o_ImageList.Create (IDB_TREECTRL_IMAGELIST, 16, 10, RGB (255, 0, 255));
	rTreeCtrl.SetImageList (&o_ImageList, TVSIL_NORMAL);

	IdentifyDoc* pDoc = GetDocument ();
	pDoc->Initialize ();
}


//
int IdentifyView::ClassNameToImageIndex (String& _rName)
{
	if (("tooltips_class32" == _rName) ||
		("VsTipWindow" == _rName) ||
		("OfficeTooltip" == _rName))
		return 2;

	if ("Button" == _rName)
		return 3;

	if (("Edit" == _rName) ||
		("VsEditPane" == _rName) ||
		("VsTextEditPane" == _rName))
		return 6;

	if ("Static" == _rName)
		return 7;

	if ("ScrollBar" == _rName)
		return 8;

	if (("ComboBox" == _rName) ||
		("ComboLBox" == _rName))
		return 9;

	if (("ListBox" == _rName) ||
		("SysListView32" == _rName))
		return 10;

	if (("CTreeView" == _rName) ||
		("CTreeCtrl" == _rName) ||
		("SysTreeView32" == _rName) ||
		("LiteTreeView32" == _rName))
		return 11;

	if ("ToolbarButton" == _rName)
		return 22;
	return 0;
}


//
int IdentifyView::UIObjectToImageIndex (UIObject* _pUIObject)
{
	if ("Button" == _pUIObject->s_ClassName)
	{
		if (0 != ((BS_3STATE|BS_AUTO3STATE|BS_AUTOCHECKBOX|BS_CHECKBOX) & _pUIObject->dw_Style))
			return 5;
		if (0 != ((BS_AUTORADIOBUTTON|BS_RADIOBUTTON) & _pUIObject->dw_Style))
			return 4;
		return 3;
	}

	return ClassNameToImageIndex (_pUIObject->s_ClassName);
}


//
void IdentifyView::InsertMenuInformation (CTreeCtrl& _rTreeCtrl, UIObject* _pUIObject, HTREEITEM _hItem)
{
	bool bSeparator = false;
	int iImageIndex = 13;
	if (0 != (MFT_MENUBARBREAK & _pUIObject->i_MenuItemType))
	{
		iImageIndex = 14;
		bSeparator = true;
	}
	if (0 != (MFT_MENUBREAK & _pUIObject->i_MenuItemType))
	{
		iImageIndex = 15;
		bSeparator = true;
	}
	if (0 != (MFT_SEPARATOR & _pUIObject->i_MenuItemType))
	{
		iImageIndex = 16;
		bSeparator = true;
	}

	String sState;
	if (0 != (MFS_CHECKED & _pUIObject->i_MenuItemState))
	{
		if (0 != (MFT_RADIOCHECK & _pUIObject->i_MenuItemType))
			iImageIndex = 18;
		else
			iImageIndex = 20;
	}
	if (0 != (MFS_UNCHECKED & _pUIObject->i_MenuItemState))
	{
		if (0 != (MFT_RADIOCHECK & _pUIObject->i_MenuItemType))
			iImageIndex = 17;
		else
			iImageIndex = 19;
	}
	if (0 != (MFS_DISABLED & _pUIObject->i_MenuItemState))
		sState += "[disabled]";
	if (0 != (MFS_GRAYED & _pUIObject->i_MenuItemState))
		sState += "[grayed]";
	if (0 != (MFS_HILITE & _pUIObject->i_MenuItemState))
		sState += "[hilite]";

	String sID;
	sID << _pUIObject->i_CommandID;

	if (false == _pUIObject->map_IdToChildren.empty ())
		iImageIndex = 21;

	String sText;
	if (false == bSeparator)
		sText << '[' << sID << "] " << _pUIObject->s_Name << ' ' << sState 
			  << " [" << (long) _pUIObject->i_MenuItemType << "]";
	HTREEITEM hItemMenu = _rTreeCtrl.InsertItem (sText, iImageIndex, iImageIndex, _hItem, TVI_LAST);
	if (false == _pUIObject->map_IdToChildren.empty ())
	{
		IdToUIObject_map_t::iterator	iteMenu;
		for (iteMenu = _pUIObject->map_IdToChildren.begin ();
			iteMenu != _pUIObject->map_IdToChildren.end ();
			++ iteMenu)
			InsertMenuInformation (_rTreeCtrl, iteMenu->second, hItemMenu);
	}
}


//
void IdentifyView::InsertWindowInformation (CTreeCtrl& _rTreeCtrl, UIObject* _pUIObject, HTREEITEM _hItem)
{
	String s;
	s << "id : " << _pUIObject->s_ID;
	_rTreeCtrl.InsertItem (s, _hItem, TVI_LAST);

	s = "";
	s << "hwnd : " << (long) _pUIObject->h_Wnd;
	_rTreeCtrl.InsertItem (s, _hItem, TVI_LAST);

	s = "";
	s << "rect : (L:" << _pUIObject->rct_Body.left <<
		",  R:" << _pUIObject->rct_Body.right << 
		",  T:" << _pUIObject->rct_Body.top <<
		",  B:" << _pUIObject->rct_Body.bottom << ")";
	_rTreeCtrl.InsertItem (s, _hItem, TVI_LAST);

	if (NULL != _pUIObject->p_MainWindow)
	{
		s = "";
		s << "main wnd : " << _pUIObject->p_MainWindow->s_Name;
		_rTreeCtrl.InsertItem (s, _hItem, TVI_LAST);
	}

	s = "";
	s << "process id : " << (long) _pUIObject->dw_ProcessID;
	_rTreeCtrl.InsertItem (s, _hItem, TVI_LAST);

	if (true == _pUIObject->b_IsVisible)
		_rTreeCtrl.InsertItem ("visible", _hItem, TVI_LAST);

	if (true == _pUIObject->b_IsChildWnd)
		_rTreeCtrl.InsertItem ("child", _hItem, TVI_LAST);

	if (true == _pUIObject->b_HasFocus)
		_rTreeCtrl.InsertItem ("has focus", _hItem, TVI_LAST);

	if (true == _pUIObject->b_IsActive)
		_rTreeCtrl.InsertItem ("active", _hItem, TVI_LAST);

	if (true == _pUIObject->b_IsForeground)
		_rTreeCtrl.InsertItem ("foreground", _hItem, TVI_LAST);

	if (true == _pUIObject->b_IsEnabledForInput)
		_rTreeCtrl.InsertItem ("input enabled", _hItem, TVI_LAST);

	s = "";
	s << "style : " << (long) _pUIObject->dw_Style;
	_rTreeCtrl.InsertItem (s, _hItem, TVI_LAST);

	s = "";
	s << "ex style : " << (long) _pUIObject->dw_ExStyle;
	_rTreeCtrl.InsertItem (s, _hItem, TVI_LAST);

	_rTreeCtrl.InsertItem (_pUIObject->GetShowState (), _hItem, TVI_LAST);
}


//
void IdentifyView::OnNMRClick (NMHDR *pNMHDR, LRESULT *pResult)
{
	CPoint posOnScreen;
	GetCursorPos (&posOnScreen);

	CPoint posOnClient = posOnScreen;
	CTreeCtrl& rTreeCtrl = CTreeView::GetTreeCtrl ();
	rTreeCtrl.ScreenToClient (&posOnClient);

	UINT uFlags;
	h_SelectedItem = rTreeCtrl.HitTest (posOnClient, &uFlags);
	if ((NULL != h_SelectedItem) && (TVHT_ONITEM & uFlags))
	   rTreeCtrl.SelectItem (h_SelectedItem);

	if (NULL == h_SelectedItem)
		return;


	CMenu menu;
	menu.CreatePopupMenu ();
	menu.AppendMenuA (MF_STRING, IDM_INTERACT_WITH_WINDOW, "Interact ...");
	menu.TrackPopupMenu (TPM_LEFTALIGN|TPM_LEFTBUTTON|TPM_RIGHTBUTTON, posOnScreen.x, posOnScreen.y, this);

	*pResult = 0;
}


//
void IdentifyView::OnInteractWithWindow ()
{
	if (NULL == h_SelectedItem)
		return;

	CTreeCtrl& rTreeCtrl = CTreeView::GetTreeCtrl ();
	UIObject* pUIObject = (UIObject*) rTreeCtrl.GetItemData (h_SelectedItem);
	if (NULL == pUIObject)
	{
		AfxMessageBox ("Cannot find UI object from selected item!");
		return;
	}

	InteractDlg dlg (this);
	dlg.s_WindowName = pUIObject->s_Name;
	dlg.s_ClassName = pUIObject->s_ClassName;

	if (IDOK != dlg.DoModal ())
		return;

	IdentifyDoc* pDoc = GetDocument ();
	pDoc->PerformAction (pUIObject->s_ID, dlg.s_Action, dlg.s_Text, NULL);
}





// IdentifyView diagnostics

#ifdef _DEBUG
void IdentifyView::AssertValid() const
{
	CTreeView::AssertValid();
}

void IdentifyView::Dump(CDumpContext& dc) const
{
	CTreeView::Dump(dc);
}

IdentifyDoc* IdentifyView::GetDocument() const // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(IdentifyDoc)));
	return (IdentifyDoc*)m_pDocument;
}
#endif //_DEBUG







