#include "stdafx.h"
#include "Identify.h"

#include "IdentifyDoc.h"
#include "RequestView.h"
#include "Error.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif


//
IMPLEMENT_DYNCREATE(RequestView, IdentifyView)
BEGIN_MESSAGE_MAP(RequestView, IdentifyView)
END_MESSAGE_MAP()


//
RequestView::RequestView()
{
	h_SelectedItem = NULL;
}

RequestView::~RequestView()
{
}


//
void RequestView::OnInitialUpdate (void)
{
	IdentifyView::OnInitialUpdate ();

	CTreeCtrl& rTreeCtrl = CTreeView::GetTreeCtrl ();

	bmp_AdditionalImages.LoadBitmap (IDB_REQUEST);
	i_RequestImageOffset = o_ImageList.Add (&bmp_AdditionalImages, RGB (255, 0, 255));
	rTreeCtrl.SetImageList (&o_ImageList, TVSIL_NORMAL);

	PopulateRequests ();
}


//
void RequestView::OnUpdate (CView* /*pSender*/, LPARAM _lHint, CObject* /*pHint*/)
{
	if (UPDATE_REQUEST_VIEW != _lHint)
		return;
	PopulateRequests ();
}


//
void RequestView::PopulateRequests (void)
{
	IdentifyDoc* pDoc = GetDocument ();
	CTreeCtrl& rTreeCtrl = IdentifyView::GetTreeCtrl ();
	rTreeCtrl.SetRedraw (FALSE);
	rTreeCtrl.DeleteAllItems ();

	{
		HTREEITEM hErrors = rTreeCtrl.InsertItem ("Window classes", TVI_ROOT, TVI_LAST);

		WindowClassToCount_map_t::iterator	ite;
		for (ite = UIObject::map_WindowClassToCount.begin ();
			ite != UIObject::map_WindowClassToCount.end ();
			++ ite)
		{
			String s;
			s << ite->first << "  [" << ite->second << "]";
			rTreeCtrl.InsertItem ((const char*)s, hErrors, TVI_LAST);
		}
	}
	if (false == Error::sdq_Errors.empty ())
	{
		int iImageIndex = i_RequestImageOffset + 6;
		HTREEITEM hErrors = rTreeCtrl.InsertItem ("Errors", iImageIndex, iImageIndex, TVI_ROOT, TVI_LAST);

		String_dq_t::iterator	ite;
		for (ite = Error::sdq_Errors.begin ();
			ite != Error::sdq_Errors.end ();
			++ ite)
			rTreeCtrl.InsertItem (*ite, hErrors, TVI_LAST);
	}
	if (false == UIObject::set_ClassesToIgnore.empty ())
	{
		int iImageIndex = i_RequestImageOffset + 5;
		HTREEITEM hWarning = rTreeCtrl.InsertItem ("Ignored classes", iImageIndex, iImageIndex, TVI_ROOT, TVI_LAST);

		String_set_t::iterator	ite;
		for (ite = UIObject::set_ClassesToIgnore.begin ();
			ite != UIObject::set_ClassesToIgnore.end ();
			++ ite)
			rTreeCtrl.InsertItem (*ite, hWarning, TVI_LAST);
	}
	if (false == UIObject::set_WindowsToIgnore.empty ())
	{
		int iImageIndex = i_RequestImageOffset + 5;
		HTREEITEM hWarning = rTreeCtrl.InsertItem ("Ignored windows", iImageIndex, iImageIndex, TVI_ROOT, TVI_LAST);

		String_set_t::iterator	ite;
		for (ite = UIObject::set_WindowsToIgnore.begin ();
			ite != UIObject::set_WindowsToIgnore.end ();
			++ ite)
			rTreeCtrl.InsertItem (*ite, hWarning, TVI_LAST);
	}

	Request_dq_t::iterator	ite;
	for (ite = Request::sdq_AllRequests.begin ();
		ite != Request::sdq_AllRequests.end ();
		++ ite)
	{
		Request* pRequest = *ite;

		int iImageIndex = i_RequestImageOffset;
		if ('!' == pRequest->c_Type)
		{
			if ("" == pRequest->s_Error)
				iImageIndex = i_RequestImageOffset + 1;
			else
				iImageIndex = i_RequestImageOffset + 2;
		}
		else if ('>' == pRequest->c_Type)
			iImageIndex = i_RequestImageOffset + 3;
		else if ('<' == pRequest->c_Type)
			iImageIndex = i_RequestImageOffset + 4;

		String sName;
		sName << '[' << pRequest->s_UIObjectID << "] " << pRequest->s_Action;
		HTREEITEM hRequest = rTreeCtrl.InsertItem (sName, iImageIndex, iImageIndex, TVI_ROOT, TVI_LAST);
		rTreeCtrl.InsertItem (pRequest->s_RawText, hRequest, TVI_LAST);
		if ("" != pRequest->s_Error)
			rTreeCtrl.InsertItem (pRequest->s_Error, hRequest, TVI_LAST);
		HTREEITEM hAffected = rTreeCtrl.InsertItem ("affected objects", hRequest, TVI_LAST);
		
		String_dq_t::iterator	iteID;
		for (iteID = pRequest->dq_AffectedUIObjectIDs.begin ();
			iteID != pRequest->dq_AffectedUIObjectIDs.end ();
			++ iteID)
		{
			UIObject* pUIObject = pDoc->p_CurrentConfiguration->Find (*iteID);
			if (NULL != pUIObject)
				InsertUIObjectToTree (rTreeCtrl, pUIObject, hAffected);
			else
			{
				String sUnknown;
				sUnknown << "Unknown object [" << *iteID << ']';
				rTreeCtrl.InsertItem (sUnknown, hAffected, TVI_LAST);
			}
		}
	}

	// rTreeCtrl.SortChildren (NULL);
	rTreeCtrl.SetRedraw (TRUE);
	INT min, max;
	rTreeCtrl.GetScrollRange (SB_VERT, &min, &max);
	rTreeCtrl.SetScrollPos (SB_VERT, max);
	rTreeCtrl.RedrawWindow ();
}


//
void RequestView::InsertUIObjectToTree (CTreeCtrl& _rTreeCtrl, UIObject* _pUIObject, HTREEITEM _hParent)
{
	if (true == _pUIObject->o_Diff.bf_Flags.b_Missing)
		return;

	String sName;
	sName += _pUIObject->s_Name + " [" + _pUIObject->s_ClassName + "]";

	int iImageIndex = UIObjectToImageIndex (_pUIObject);
	if (0 == iImageIndex)
		iImageIndex = 1;
	if (("ToolbarButton" == _pUIObject->s_ClassName) &&
		(_pUIObject->rct_Body.IsRectEmpty ()))
		++ iImageIndex;

	HTREEITEM hItem = _rTreeCtrl.InsertItem (sName, iImageIndex, iImageIndex, _hParent, TVI_LAST);
	_rTreeCtrl.SetItemData (hItem, (DWORD_PTR)_pUIObject);

	HTREEITEM hProperties = _rTreeCtrl.InsertItem ("Properties", hItem, TVI_LAST);
	InsertWindowInformation (_rTreeCtrl, _pUIObject, hProperties);

	if (false == _pUIObject->map_CommandIdToMenuItems.empty ())
	{
		HTREEITEM hItemMenu = _rTreeCtrl.InsertItem ("Menu", hItem, TVI_LAST);
		IdToUIObject_map_t::iterator	iteMenu;
		for (iteMenu = _pUIObject->map_CommandIdToMenuItems.begin ();
			iteMenu != _pUIObject->map_CommandIdToMenuItems.end ();
			++ iteMenu)
			InsertMenuInformation (_rTreeCtrl, iteMenu->second, hItemMenu);
	}

	IdToUIObject_map_t::iterator	ite;
	for (ite = _pUIObject->map_IdToChildren.begin ();
		ite != _pUIObject->map_IdToChildren.end ();
		++ ite)
	{
		InsertUIObjectToTree (_rTreeCtrl, ite->second, hItem);
	}
}





// RequestView diagnostics

#ifdef _DEBUG
void RequestView::AssertValid() const
{
	IdentifyView::AssertValid();
}

void RequestView::Dump(CDumpContext& dc) const
{
	IdentifyView::Dump(dc);
}

IdentifyDoc* RequestView::GetDocument() const // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(IdentifyDoc)));
	return (IdentifyDoc*)m_pDocument;
}
#endif //_DEBUG








