#include "stdafx.h"
#include "Socket.h"


// ======================================================================
ServerSocket::ServerSocket (void)
{
	p_Callback = NULL;
}

	
ServerSocket::~ServerSocket (void)
{
	ConnectionSocket_set_t::iterator	ite;
	for (ite = set_ConnectionSockets.begin ();
		ite != set_ConnectionSockets.end ();
		++ ite)
	{
		ConnectionSocket* pConnection = *ite;
		delete pConnection;
	}

	set_ConnectionSockets.clear ();
}


//
void ServerSocket::RemoveConnectionSocket (ConnectionSocket* _pSocket)
{
	set_ConnectionSockets.erase (_pSocket);
}


//
void ServerSocket::SetCallback (SocketCallback* _pCallback)
{
	p_Callback = _pCallback;
}


//
void ServerSocket::OnAccept (int nErrorCode)
{
	if (NULL == p_Callback)
		return;

	ConnectionSocket* pSocket = new ConnectionSocket (p_Callback, this);
	set_ConnectionSockets.insert (pSocket);

	CAsyncSocket::Accept (*pSocket);
	// TRACE ( "======================= Connection received =======================\n");
}



// ======================================================================
ConnectionSocket::ConnectionSocket (SocketCallback* _pCallback, ServerSocket* _pServerSocket)
{
	b_Connected = false;
	p_Callback = _pCallback;
	p_ServerSocket = _pServerSocket;
	if (NULL == p_ServerSocket)
		Create ();
}


//
ConnectionSocket::~ConnectionSocket (void)
{
	Close ();
	p_Callback->OnDestroySocket ();
}

//
void ConnectionSocket::OnConnect (int nErrorCode)
{
	if (0 != nErrorCode)
		return;
	b_Connected = true;
	p_Callback->OnConnect ();
}


//
void ConnectionSocket::OnClose (int nErrorCode)
{
	b_Connected = false;
	p_Callback->OnDisconnect ();

	if (NULL != p_ServerSocket)
		p_ServerSocket->RemoveConnectionSocket (this);
	delete this;
}



//
bool ConnectionSocket::Send (const char* _pBuf)
{
	int iLength = strlen (_pBuf);
	while (true)
	{
		if (SOCKET_ERROR != CAsyncSocket::Send (_pBuf, iLength))
			return true;
		int iError = CAsyncSocket::GetLastError ();
		if ((WSAEINPROGRESS != iError) &&
			(WSAENOBUFS != iError) &&
			(WSAEWOULDBLOCK != iError))
			break;
		Sleep (100);
	}
	b_Connected = false;
	return false;
}



//
void ConnectionSocket::OnReceive (int nErrorCode)
{
	DWORD iAvailableData;
	IOCtl (FIONREAD, &iAvailableData);
	char* zData = new char [iAvailableData + 10];
	int iLength = Receive (zData, iAvailableData + 10);
	zData [iLength] = '\0';

	p_Callback->OnReceive (zData, iLength);

	delete[] zData;
}


//
bool ConnectionSocket::IsConnected (void)
{
	return b_Connected;
}
