#include "stdafx.h"
#include "SystemMonitor.h"


//
SystemMonitor::SystemMonitor (void)
{
	memset (d_Processes, 0, sizeof (double) * 9);
	memset (d_ProcessorUsage, 0, sizeof (double) * 9);
	memset (d_AvailableMemory, 0, sizeof (double) * 9);
	memset (d_LogicalDiskBytes, 0, sizeof (double) * 9);
	memset (d_NetworkBytes, 0, sizeof (double) * 9);
	i_MeasurementIndex = 9;
}

SystemMonitor::~SystemMonitor (void)
{
	// ResetAverage ();
	PdhRemoveCounter (hct_Processes);
	PdhRemoveCounter (hct_ProcessorUsage);
	PdhRemoveCounter (hct_AvailableMemory);
	PdhRemoveCounter (hct_LogicalDiskBytes);
	PdhRemoveCounter (hct_NetworkBytes);
	PdhCloseQuery (h_Query);
}


//
bool SystemMonitor::Init (void)
{
	PDH_STATUS hStatus;
	
	hStatus = PdhOpenQuery (NULL, 0, &h_Query);
	if (ERROR_SUCCESS != hStatus)
	{
		TRACE ("PdhOpenQuery failed.");
		return false;
	}

	if (ERROR_SUCCESS != PdhAddCounter (h_Query, "\\System\\Processes", 0, &hct_Processes))
	{
		PdhCloseQuery (h_Query);
		AfxMessageBox ("Failed to initialize 'processes' counter.");
		return false;
	}
	if (ERROR_SUCCESS != PdhAddCounter (h_Query, "\\Processor(_Total)\\% Processor Time", 0, &hct_ProcessorUsage))
	{
		PdhCloseQuery (h_Query);
		AfxMessageBox ("Failed to initialize 'CPU' counter.");
		return false;
	}
	if (ERROR_SUCCESS != PdhAddCounter (h_Query, "\\Memory\\Available KBytes", 0, &hct_AvailableMemory))
	{
		PdhCloseQuery (h_Query);
		AfxMessageBox ("Failed to initialize 'memory' counter.");
		return false;
	}
	if (ERROR_SUCCESS != PdhAddCounter (h_Query, "\\PhysicalDisk(_Total)\\Disk Bytes/sec", 0, &hct_LogicalDiskBytes))
	{
		PdhCloseQuery (h_Query);
		AfxMessageBox ("Failed to initialize 'disk' counter.");
		return false;
	}
	if (ERROR_SUCCESS != PdhAddCounter (h_Query, "\\Network Interface(*)\\Bytes Total/sec", 0, &hct_NetworkBytes))
	{
		PdhCloseQuery (h_Query);
		AfxMessageBox ("Failed to initialize 'network' counter.");
		return false;
	}

	memset (d_Processes, 0, sizeof (double) * 9);
	memset (d_ProcessorUsage, 0, sizeof (double) * 9);
	memset (d_AvailableMemory, 0, sizeof (double) * 9);
	memset (d_LogicalDiskBytes, 0, sizeof (double) * 9);
	memset (d_NetworkBytes, 0, sizeof (double) * 9);
	i_MeasurementIndex = 9;

	PdhCollectQueryData (h_Query);
	return true;
}


//
/*
void SystemMonitor::ResetAverage (void)
{
	d_ProcessesTotal = 0;
	d_ProcessorUsageTotal = 0;
	d_AvailableMemoryTotal = 0;
	d_LogicalDiskBytesTotal = 0;
	d_NetworkBytesTotal = 0;
	l_Measurements = 0;
}
*/


//
bool SystemMonitor::MeasureAverageLoad (void)
{
	PdhCollectQueryData (h_Query);

	DWORD dwType;
	PDH_FMT_COUNTERVALUE oValue;

	double dProcesses;	
	if (ERROR_SUCCESS != PdhGetFormattedCounterValue (hct_Processes, PDH_FMT_DOUBLE, &dwType, &oValue))
		return false;
	dProcesses = oValue.doubleValue;

	double dProcessorUsage;	
	if (ERROR_SUCCESS != PdhGetFormattedCounterValue (hct_ProcessorUsage, PDH_FMT_DOUBLE, &dwType, &oValue))
		return false;
	dProcessorUsage = oValue.doubleValue;

	double dAvailableMemory;	
	if (ERROR_SUCCESS != PdhGetFormattedCounterValue (hct_AvailableMemory, PDH_FMT_DOUBLE, &dwType, &oValue))
		return false;
	dAvailableMemory = oValue.doubleValue;

	double dDiskBytes;
	if (ERROR_SUCCESS != PdhGetFormattedCounterValue (hct_LogicalDiskBytes, PDH_FMT_DOUBLE, &dwType, &oValue))
		return false;
	dDiskBytes = oValue.doubleValue;

	double dNetworkBytes;
	if (ERROR_SUCCESS != PdhGetFormattedCounterValue (hct_NetworkBytes, PDH_FMT_DOUBLE, &dwType, &oValue))
		return false;
	dNetworkBytes = oValue.doubleValue;

	//++ l_Measurements;
	//d_ProcessesTotal += dProcesses;
	//d_ProcessorUsageTotal += dProcessorUsage;
	//d_AvailableMemoryTotal += dAvailableMemory;
	//d_LogicalDiskBytesTotal += dDiskBytes;
	//d_NetworkBytesTotal += dNetworkBytes;
	d_Processes [i_MeasurementIndex % 9] = dProcesses;
	d_ProcessorUsage [i_MeasurementIndex % 9] = dProcessorUsage;
	d_AvailableMemory [i_MeasurementIndex % 9] = dAvailableMemory;
	d_LogicalDiskBytes [i_MeasurementIndex % 9] = dDiskBytes;
	d_NetworkBytes [i_MeasurementIndex % 9] = dNetworkBytes;
	++ i_MeasurementIndex;
	if (i_MeasurementIndex >= 18)
		i_MeasurementIndex = 9;

	return true;
}


//
int SystemMonitor::CheckCurrentLoad (void)
{
	if (false == MeasureAverageLoad ())
		return 0;

	// compare averages ...
	int iRet = CompareAverages ();
	//TRACE ("\n");
	return iRet;
}


//
int SystemMonitor::CompareAverages (void)
{
	if (0 <= CompareAverage ("processes", d_Processes))
		return 1;
	if (0 <= CompareAverage ("cpu", d_ProcessorUsage))
		return 1;
	if (0 <= CompareAverage ("memory", d_AvailableMemory))
		return 1;
	//if (0 <= CompareAverage ("disk", d_LogicalDiskBytes))
	//	return 1;
	//if (0 <= CompareAverage ("network", d_NetworkBytes))
	//	return 1;
	return -1;
}


//
int SystemMonitor::CompareAverage (const char* _zName, double* _dArray)
{
	double dOldAverage = (_dArray [(i_MeasurementIndex - 6) % 9] + 
						  _dArray [(i_MeasurementIndex - 5) % 9] +
						  _dArray [(i_MeasurementIndex - 4) % 9]) / (double) 3;

	double dNewAverage = (_dArray [(i_MeasurementIndex - 3) % 9] + 
						  _dArray [(i_MeasurementIndex - 2) % 9] +
						  _dArray [(i_MeasurementIndex - 1) % 9]) / (double) 3;

	//TRACE ("%s: %f -- %f\n", _zName, 1.1 * dOldAverage, dNewAverage);
	return (dNewAverage <= 1.1 * dOldAverage)? -1 : 1;
}


//
/*
bool SystemMonitor::CheckCurrentLoad (void)
{
	PdhCollectQueryData (h_Query);

	DWORD dwType;
	PDH_FMT_COUNTERVALUE oValue;

	double dProcesses;	
	if (ERROR_SUCCESS != PdhGetFormattedCounterValue (hct_Processes, PDH_FMT_DOUBLE, &dwType, &oValue))
		return false;
	dProcesses = oValue.doubleValue;

	double dProcessorUsage;	
	if (ERROR_SUCCESS != PdhGetFormattedCounterValue (hct_ProcessorUsage, PDH_FMT_DOUBLE, &dwType, &oValue))
		return false;
	dProcessorUsage = oValue.doubleValue;

	double dAvailableMemory;	
	if (ERROR_SUCCESS != PdhGetFormattedCounterValue (hct_AvailableMemory, PDH_FMT_DOUBLE, &dwType, &oValue))
		return false;
	dAvailableMemory = oValue.doubleValue;

	double dDiskBytes;
	if (ERROR_SUCCESS != PdhGetFormattedCounterValue (hct_LogicalDiskBytes, PDH_FMT_DOUBLE, &dwType, &oValue))
		return false;
	dDiskBytes = oValue.doubleValue;

	double dNetworkBytes;
	if (ERROR_SUCCESS != PdhGetFormattedCounterValue (hct_LogicalDiskBytes, PDH_FMT_DOUBLE, &dwType, &oValue))
		return false;
	dNetworkBytes = oValue.doubleValue;

	double dAverage;
	dAverage = d_ProcessesTotal / (double) l_Measurements;
	if (dProcesses > 1.2 * dAverage)
		return false;
	dAverage = d_ProcessorUsageTotal / (double) l_Measurements;
	if (dProcessorUsage > 1.2 * dAverage)
		return false;
	dAverage = d_AvailableMemoryTotal / (double) l_Measurements;
	if (dAvailableMemory > 1.2 * dAverage)
		return false;
	dAverage = d_LogicalDiskBytesTotal / (double) (l_Measurements - 1);
	if (dDiskBytes > 1.2 * dAverage)
		return false;
	dAverage = d_NetworkBytesTotal / (double) (l_Measurements - 1);
	if (dNetworkBytes > 1.2 * dAverage)
		return false;
	
	return true;
}
*/