#include "StdAfx.h"
#include <nlp_filesystem.h>
#include "UIObject.h"
#include "Error.h"


#ifndef _DEBUG
	#define assert	;//
#endif


Request_dq_t				Request::sdq_AllRequests;
WindowClassToCount_map_t	UIObject::map_WindowClassToCount;
SystemConfiguration*		UIObject::p_CurrentSystemConfig = NULL;
ProcessID_set_t				UIObject::set_ProcessIDs;
DWORD						UIObject::dw_MyProcessID = -1;
String_set_t				UIObject::set_ClassesToIgnore;
String_set_t				UIObject::set_WindowsToIgnore;
String_map_t				UIObject::map_IdToOldName;
bool						UIObject::b_StartButtonIdentified = false;
bool						UIObject::b_StartMenuIdentified = false;
long						UIObject::sl_ObjectCount = 0;
UIObject_dq_t				UIObject::sdq_TemporaryObjects;

//
Request::Request (void)
{
	c_Type = ' ';
	Request::sdq_AllRequests.push_back (this);
}

Request::~Request (void)
{
}

void Request::Clear (void)
{
	Request_dq_t::iterator	ite;
	for (ite = Request::sdq_AllRequests.begin ();
		ite != Request::sdq_AllRequests.end ();
		++ ite)
		delete *ite;

	Request::sdq_AllRequests.clear ();
}

//
SystemConfiguration::SystemConfiguration (void)
{
}


//
SystemConfiguration::~SystemConfiguration (void)
{
	map_IdToUIObject.clear ();
	mmap_NameToUIObject.clear ();
	dq_AllUIObjects.clear ();

	IdToUIObject_map_t::iterator	ite;
	for (ite = map_IdToTopmostUIObjects.begin ();
		ite != map_IdToTopmostUIObjects.end ();
		++ ite)
		delete ite->second;

	map_IdToTopmostUIObjects.clear ();
}


//
bool SystemConfiguration::IsChanging (SystemConfiguration& _rOther, ProcessID_set_t& _rsetProcessesToIgnore)
{
	bool bIsDifferent = false;
	{
		IdToUIObject_map_t::iterator	iteThis;
		for (iteThis = map_IdToTopmostUIObjects.begin ();
			iteThis != map_IdToTopmostUIObjects.end ();
			++ iteThis)
		{
			UIObject* pThisObject = iteThis->second;
			if (_rsetProcessesToIgnore.end () != _rsetProcessesToIgnore.find (pThisObject->dw_ProcessID))
				continue;

			IdToUIObject_map_t::iterator	iteOther;
			iteOther = _rOther.map_IdToTopmostUIObjects.find (iteThis->first);
			if (iteOther == _rOther.map_IdToTopmostUIObjects.end ())
			{
				pThisObject->o_Diff.bf_Flags.b_New = true;
				if (true == pThisObject->b_IsVisible)
					pThisObject->o_Diff.bf_Flags.b_IsVisible = true;
				if (true == pThisObject->b_HasFocus)
					pThisObject->o_Diff.bf_Flags.b_HasFocus = true;
				if (true == pThisObject->b_IsActive)
					pThisObject->o_Diff.bf_Flags.b_IsActive = true;
				if (true == pThisObject->b_IsForeground)
					pThisObject->o_Diff.bf_Flags.b_IsForeground = true;
				if (true == pThisObject->b_IsEnabledForInput)
					pThisObject->o_Diff.bf_Flags.b_IsEnabledForInput = true;
				bIsDifferent = true;
			}
			else
				bIsDifferent |= pThisObject->Diff (*iteOther->second, false);
		}
	}
	{
		IdToUIObject_map_t::iterator	iteOther;
		for (iteOther = _rOther.map_IdToTopmostUIObjects.begin ();
			iteOther != _rOther.map_IdToTopmostUIObjects.end ();
			++ iteOther)
		{
			UIObject* pOtherObject = iteOther->second;
			if (_rsetProcessesToIgnore.end () != _rsetProcessesToIgnore.find (pOtherObject->dw_ProcessID))
				continue;

			IdToUIObject_map_t::iterator	iteThis;
			iteThis = map_IdToTopmostUIObjects.find (iteOther->first);
			if (iteThis == map_IdToTopmostUIObjects.end ())
			{
				UIObject* pOtherChild = iteOther->second;
				UIObject* pThisChild = pOtherChild->Clone (NULL);
				pThisChild->o_Diff.bf_Flags.b_Missing = true;
				map_IdToTopmostUIObjects.insert (make_pair (pThisChild->s_ID, pThisChild));
				bIsDifferent = true;
			}
		}
	}

	return bIsDifferent;
}


//
bool SystemConfiguration::Diff (SystemConfiguration& _rOther, bool _bPreserveLastDiff)
{
	bool bIsDifferent = false;
	{
		IdToUIObject_map_t::iterator	iteThis;
		for (iteThis = map_IdToTopmostUIObjects.begin ();
			iteThis != map_IdToTopmostUIObjects.end ();
			++ iteThis)
		{
			UIObject* pThisObject = iteThis->second;
			if (false == _bPreserveLastDiff)
				pThisObject->o_Diff.ui_Difference = 0;

			IdToUIObject_map_t::iterator	iteOther;
			iteOther = _rOther.map_IdToTopmostUIObjects.find (iteThis->first);
			if (iteOther == _rOther.map_IdToTopmostUIObjects.end ())
			{
				pThisObject->o_Diff.bf_Flags.b_New = true;
				if (true == pThisObject->b_IsVisible)
					pThisObject->o_Diff.bf_Flags.b_IsVisible = true;
				if (true == pThisObject->b_HasFocus)
					pThisObject->o_Diff.bf_Flags.b_HasFocus = true;
				if (true == pThisObject->b_IsActive)
					pThisObject->o_Diff.bf_Flags.b_IsActive = true;
				if (true == pThisObject->b_IsForeground)
					pThisObject->o_Diff.bf_Flags.b_IsForeground = true;
				if (true == pThisObject->b_IsEnabledForInput)
					pThisObject->o_Diff.bf_Flags.b_IsEnabledForInput = true;
				bIsDifferent = true;
			}
			else
				bIsDifferent |= pThisObject->Diff (*iteOther->second, _bPreserveLastDiff);
		}
	}
	{
		IdToUIObject_map_t::iterator	iteOther;
		for (iteOther = _rOther.map_IdToTopmostUIObjects.begin ();
			iteOther != _rOther.map_IdToTopmostUIObjects.end ();
			++ iteOther)
		{
			UIObject* pOtherChild = iteOther->second;
			if (true == pOtherChild->o_Diff.bf_Flags.b_Missing)
				continue;

			IdToUIObject_map_t::iterator	iteThis;
			iteThis = map_IdToTopmostUIObjects.find (iteOther->first);
			if (iteThis == map_IdToTopmostUIObjects.end ())
			{
				UIObject* pThisChild = pOtherChild->Clone (NULL);
				pThisChild->o_Diff.bf_Flags.b_Missing = true;
				map_IdToTopmostUIObjects.insert (make_pair (pThisChild->s_ID, pThisChild));
				bIsDifferent = true;
			}
		}
	}

	return bIsDifferent;
}


//
bool SystemConfiguration::RollbackStartMenu (SystemConfiguration& _rOriginal, String& _rError)
{
	IdToUIObject_map_t::iterator	ite;
	for (ite = map_IdToTopmostUIObjects.begin ();
		ite != map_IdToTopmostUIObjects.end ();
		++ ite)
	{
		UIObject* pObject = ite->second;
		if (pObject->o_Diff.bf_Flags.b_Missing)
			continue;

		UIObject* pOriginal = NULL;
		IdToUIObject_map_t::iterator	iteOriginal;
		iteOriginal = _rOriginal.map_IdToTopmostUIObjects.find (pObject->s_ID);
		if (iteOriginal != _rOriginal.map_IdToTopmostUIObjects.end ())
			pOriginal = iteOriginal->second;

		pObject->RollbackStartMenu (pOriginal, _rError);
	}
	return true;
}


//
bool SystemConfiguration::Rollback (SystemConfiguration& _rOriginal, String& _rError)
{
	bool bComplete = true;

	IdToUIObject_map_t::iterator	ite;
	for (ite = map_IdToTopmostUIObjects.begin ();
		ite != map_IdToTopmostUIObjects.end ();
		++ ite)
	{
		UIObject* pObject = ite->second;

		if (pObject->o_Diff.bf_Flags.b_Missing)
		{
			bComplete = false;
			_rError << "\tmissing window : [" << pObject->s_Name << "] (" << pObject->s_ClassName << ")\n";
		}
		else
		{
			UIObject* pOriginal = NULL;
			IdToUIObject_map_t::iterator	iteOriginal;
			iteOriginal = _rOriginal.map_IdToTopmostUIObjects.find (pObject->s_ID);
			if (iteOriginal != _rOriginal.map_IdToTopmostUIObjects.end ())
				pOriginal = iteOriginal->second;

			pObject->Rollback (pOriginal, _rError);
		}
	}

	return bComplete;
}


//
bool SystemConfiguration::Find (const char* _zName, UIObject_dq_t* _pdqUIObjects)
{
	pair <NameToUIObject_mmap_t::iterator, NameToUIObject_mmap_t::iterator>	pairResults;
	pairResults = mmap_NameToUIObject.equal_range (_zName);

	NameToUIObject_mmap_t::iterator	ite;
	for (ite = pairResults.first; ite != pairResults.second; ++ ite)
		_pdqUIObjects->push_back (ite->second);

	return (pairResults.first != pairResults.second);
}


//
void SystemConfiguration::ListAllObjects (UIObject_dq_t* _pdqUIObjects)
{
	NameToUIObject_mmap_t::iterator ite;
	for (ite = mmap_NameToUIObject.begin ();
		ite != mmap_NameToUIObject.end ();
		++ ite)
	{
		UIObject* pObject = ite->second;

		bool bVisible = pObject->b_IsVisible;
		if (NULL != pObject->p_ActivationWindow)
			bVisible &= pObject->p_ActivationWindow->b_IsVisible;

		bool bInvisibleRect = ((-1 == pObject->rct_Body.left) &&
								(-1 == pObject->rct_Body.right) &&
								(-1 == pObject->rct_Body.top) &&
								(-1 == pObject->rct_Body.bottom));
		if (NULL != pObject->p_Parent)
		{
			bInvisibleRect |= ((-1 == pObject->p_Parent->rct_Body.left) &&
									(-1 == pObject->p_Parent->rct_Body.right) &&
									(-1 == pObject->p_Parent->rct_Body.top) &&
									(-1 == pObject->p_Parent->rct_Body.bottom));
		}
		if (NULL != pObject->p_ActivationWindow)
		{
			bInvisibleRect |= ((-1 == pObject->p_ActivationWindow->rct_Body.left) &&
									(-1 == pObject->p_ActivationWindow->rct_Body.right) &&
									(-1 == pObject->p_ActivationWindow->rct_Body.top) &&
									(-1 == pObject->p_ActivationWindow->rct_Body.bottom));
		}
		if ((false == bVisible) || (true == bInvisibleRect))
			continue;

		if ((UIOBJECT__MENU_ITEM == pObject->c_ObjectType) &&
			(MFS_GRAYED == pObject->i_MenuItemState))
			continue;
		//if (("Static" == pObject->s_ClassName) || ("groupbox" == pObject->s_ClassName))
		//	continue;

		if ("DV2ControlHost" == pObject->p_ActivationWindow->s_ClassName)
		{
			if (("Log Off" == pObject->s_Name) ||
				("Turn Off Computer" == pObject->s_Name))
				continue;
		}
		if ("Static" == pObject->s_ClassName)
		{
			pObject->GetNeighbouringSibling (*_pdqUIObjects);
			continue;
		}
		//else if ("groupbox" == pObject->s_ClassName)
		//	pObject->GetEnclosedSibling (*_pdqUIObjects);
		_pdqUIObjects->push_back (pObject);
	}
}


//
bool SystemConfiguration::FindApproximate (const char* _zLowercaseName, UIObject_dq_t* _pdqUIObjects)
{
	NameToUIObject_mmap_t::iterator ite;
	for (ite = mmap_NameToUIObject.begin ();
		ite != mmap_NameToUIObject.end ();
		++ ite)
	{
		UIObject* pObject = ite->second;
		bool bVisible = pObject->b_IsVisible;
		if (NULL != pObject->p_ActivationWindow)
			bVisible &= pObject->p_ActivationWindow->b_IsVisible;


		bool bInvisibleRect = ((-1 == pObject->rct_Body.left) &&
								(-1 == pObject->rct_Body.right) &&
								(-1 == pObject->rct_Body.top) &&
								(-1 == pObject->rct_Body.bottom));
		if (NULL != pObject->p_Parent)
		{
			bInvisibleRect |= ((-1 == pObject->p_Parent->rct_Body.left) &&
									(-1 == pObject->p_Parent->rct_Body.right) &&
									(-1 == pObject->p_Parent->rct_Body.top) &&
									(-1 == pObject->p_Parent->rct_Body.bottom));
		}
		bInvisibleRect |= ((-1 == pObject->p_ActivationWindow->rct_Body.left) &&
								(-1 == pObject->p_ActivationWindow->rct_Body.right) &&
								(-1 == pObject->p_ActivationWindow->rct_Body.top) &&
								(-1 == pObject->p_ActivationWindow->rct_Body.bottom));

		if ((false == bVisible) || (true == bInvisibleRect))
			continue;

		if (UIOBJECT__MENU_ITEM == pObject->c_ObjectType)
			if (MFS_GRAYED & pObject->i_MenuItemState)
				continue;

		if (true == pObject->s_LowercaseName.Has (_zLowercaseName))
		{
			if ("DV2ControlHost" == pObject->p_ActivationWindow->s_ClassName)
			{
				if (("Log Off" == pObject->s_Name) ||
					("Turn Off Computer" == pObject->s_Name))
					continue;
			}
			if ("Static" == pObject->s_ClassName)
				pObject->GetNeighbouringSibling (*_pdqUIObjects);
			//else if ("groupbox" == pObject->s_ClassName)
			//	pObject->GetEnclosedSibling (*_pdqUIObjects);
			else
				_pdqUIObjects->push_back (pObject);
		}
	}

	return (false == _pdqUIObjects->empty ());
}


//
UIObject* SystemConfiguration::Find (const char* _zID)
{
	IdToUIObject_map_t::iterator	ite;
	ite = map_IdToUIObject.find (_zID);
	if (ite == map_IdToUIObject.end ())
		return NULL;
	return ite->second;
}


//
UIObject* SystemConfiguration::Find (HWND _hWnd)
{
	HWndToUIObject_map_t::iterator	ite;
	ite = map_HWndToUIObject.find (_hWnd);
	if (ite == map_HWndToUIObject.end ())
		return NULL;
	return ite->second;
}


//
void SystemConfiguration::CreateWindowTree (void)
{
	{
		UIObject_dq_t::iterator	ite;
		for (ite = dq_AllUIObjects.begin ();
			ite != dq_AllUIObjects.end ();
			++ ite)
		{
			UIObject* pUIObject = *ite;
			// tree view items have already been linked up with their parent items ...
			if (UIOBJECT__TREEVIEW_ITEM == pUIObject->c_ObjectType)
				continue;
			if (NULL == pUIObject->h_ParentWnd)
			{
				SystemConfiguration::map_IdToTopmostUIObjects.insert (make_pair (pUIObject->s_ID, pUIObject));
				continue;
			}
			UIObject* pParent = SystemConfiguration::Find (pUIObject->h_ParentWnd);
			if (NULL == pParent)
			{
				SystemConfiguration::map_IdToTopmostUIObjects.insert (make_pair (pUIObject->s_ID, pUIObject));
				continue;
			}

			pUIObject->p_Parent = pParent;
			pParent->AddChild (pUIObject);
		}
	}

	{
		//NameToUIObject_mmap_t::iterator ite;
		//for (ite = mmap_NameToUIObject.begin ();
		//	ite != mmap_NameToUIObject.end ();
		//	++ ite)

		IdToUIObject_map_t::iterator	ite;
		for (ite = map_IdToUIObject.begin ();
			ite != map_IdToUIObject.end ();
			++ ite)
		{
			UIObject* pUIObject = ite->second;
			//if ("Popup Menu" == pUIObject->s_ClassName)
			//	TRACE ("\n");
			UIObject* pTopParent = pUIObject;
			while (NULL != pTopParent->p_Parent)
			{
				bool bActivationWindow = false;
				if (UIOBJECT__DIALOG == pTopParent->c_ObjectType)
					bActivationWindow = true;
				if ("DV2ControlHost" == pTopParent->s_ClassName)
					bActivationWindow = true;
				if ((NULL == pUIObject->p_ActivationWindow) &&
					(true == bActivationWindow))
					pUIObject->p_ActivationWindow = pTopParent;
				pTopParent = pTopParent->p_Parent;
			}
			pUIObject->p_MainWindow = pTopParent;
			if (NULL == pUIObject->p_ActivationWindow)
				pUIObject->p_ActivationWindow = pTopParent;
			if ((false == UIObject::b_StartMenuIdentified) && ("SysPager" == pUIObject->s_ClassName))
			{
				pUIObject->p_MainWindow->b_StartMenu = true;
				UIObject::b_StartMenuIdentified = true;
			}
		}
	}
	ASSERT (true == UIObject::b_StartMenuIdentified);

	{
		p_ForegroundObject = NULL;

		IdToUIObject_map_t::iterator	ite;
		for (ite = map_IdToTopmostUIObjects.begin ();
			ite != map_IdToTopmostUIObjects.end ();
			++ ite)
		{
			UIObject* pObject = ite->second;
			if (true == pObject->b_IsForeground)
				p_ForegroundObject = pObject;
		}
	}

	{
		p_FocusedObject = NULL;

		TRACE ("checking for focused object\n");
		IdToUIObject_map_t::iterator	ite;
		for (ite = map_IdToUIObject.begin ();
			ite != map_IdToUIObject.end ();
			++ ite)
		{
			UIObject* pObject = ite->second;
			if (true == pObject->b_HasFocus)
			{
				p_FocusedObject = pObject;
				TRACE ("focused : %s [%s] %s\n", (const char*)pObject->s_ID, (const char*)pObject->s_ClassName, (const char*)pObject->s_Name);
			}
			if (true == pObject->b_IsForeground)
			{
				TRACE ("fg : %s [%s] %s\n", (const char*)pObject->s_ID, (const char*)pObject->s_ClassName, (const char*)pObject->s_Name);
			}
			if (true == pObject->b_IsActive)
			{
				TRACE ("active : %s [%s] %s\n", (const char*)pObject->s_ID, (const char*)pObject->s_ClassName, (const char*)pObject->s_Name);
			}
		}
	}
}



//
String& operator<< (String& _rLeft, const UIObjectChange& _oDiff)
{
	_rLeft	<< "~cld:" << ((true == _oDiff.bf_Flags.b_ChildIsDifferent)? 1 : 0)
			<< " ~nme:" << ((true == _oDiff.bf_Flags.b_Name)? 1 : 0)
			<< " ~mnu:" << ((true == _oDiff.bf_Flags.b_MenuItemState)? 1 : 0)
			<< " ~rct:" << ((true == _oDiff.bf_Flags.b_RectBody)? 1 : 0)
			<< " ~vsb:" << ((true == _oDiff.bf_Flags.b_IsVisible)? 1 : 0)
			<< " ~fcs:" << ((true == _oDiff.bf_Flags.b_HasFocus)? 1 : 0)
			<< " ~act:" << ((true == _oDiff.bf_Flags.b_IsActive)? 1 : 0)
			<< " ~fgr:" << ((true == _oDiff.bf_Flags.b_IsForeground)? 1 : 0)
			<< " ~ien:" << ((true == _oDiff.bf_Flags.b_IsEnabledForInput)? 1 : 0)
			<< " ~sty:" << ((true == _oDiff.bf_Flags.b_Style)? 1 : 0)
			<< " ~exs:" << ((true == _oDiff.bf_Flags.b_ExStyle)? 1 : 0)
			<< " ~sst:" << ((true == _oDiff.bf_Flags.b_ShowState)? 1 : 0)
			<< " ~new:" << ((true == _oDiff.bf_Flags.b_New)? 1 : 0)
			<< " ~old:" << ((true == _oDiff.bf_Flags.b_Missing)? 1 : 0);
	return _rLeft;
}


//
void UIObject::Init (void)
{
	h_Wnd = NULL;
	h_ParentWnd = NULL;
	h_Menu = NULL;
	p_Parent = NULL;
	p_MainWindow = NULL;
	p_ActivationWindow = NULL;

	rct_Body.SetRectEmpty ();

	i_MenuItemState = 0;
	dw_Style = 0;
	dw_ExStyle = 0;
	ui_ShowState = 0;

	b_StartMenu = false;
	b_IsVisible = false;
	b_IsSelfVisible = false;
	b_HasFocus = false;
	b_IsActive = false;
	b_IsForeground = false;
	b_IsEnabledForInput = false;
	b_IsChildWnd = false;
	memset (&o_Diff, 0, sizeof (o_Diff));

	c_ObjectType = UIOBJECT__UNKNOWN;
}


//
UIObject::UIObject (void)
{
	Init ();
}


//
UIObject::UIObject (CWnd* _pWnd)
{
	Init ();

	// Ordinary window ...
	dw_ProcessID = 0;
	GetWindowThreadProcessId (_pWnd->GetSafeHwnd (), &dw_ProcessID);

	c_ObjectType = UIOBJECT__WINDOW;
	ExtractWindowInformation (_pWnd);

	s_Name.Replace ("&&", "\x1");
	s_Name.Replace ("&", "");
	s_Name.Replace ("\x1", "&");
	s_LowercaseName = s_Name;
	s_LowercaseName.LowerCase ();
	//s_ID << (long)h_ParentWnd << '-' << (long)h_Wnd << '-' << c_ObjectType << '-' << i_CommandID << '-' << sl_ObjectCount;
	s_ID << (long)h_ParentWnd << '-' << (long)h_Wnd << '-' << c_ObjectType << '-' << i_CommandID;
	++ sl_ObjectCount;

	UIObject::set_ProcessIDs.insert (dw_ProcessID);
	UIObject::p_CurrentSystemConfig->dq_AllUIObjects.push_back (this);
	UIObject::p_CurrentSystemConfig->map_IdToUIObject.insert (make_pair (s_ID, this));
	UIObject::p_CurrentSystemConfig->map_HWndToUIObject.insert (make_pair (h_Wnd, this));
	UIObject::p_CurrentSystemConfig->mmap_NameToUIObject.insert (make_pair (s_Name, this));
}


//
UIObject::UIObject (UIObject* _pParentWnd, HWND _hToolbarWnd, int _iButtonIndex, char _cObjectType)
{
	Init ();

	// toolbar button or listview item ...
	h_ParentWnd = _hToolbarWnd;
	p_Parent = _pParentWnd;
	ui_ShowState = _pParentWnd->ui_ShowState;
	dw_ProcessID = _pParentWnd->dw_ProcessID;

	c_ObjectType = _cObjectType;
	if (UIOBJECT__TOOLBAR_BUTTON == c_ObjectType)
	{
		ExtractToolbarButtonInformation (_hToolbarWnd, _iButtonIndex);
		i_CommandID = _iButtonIndex;
		if (rct_Body.IsRectEmpty ())
		{
			delete this;
			return;
		}
	}
	else if (UIOBJECT__LISTVIEW_BUTTON == c_ObjectType)
		ExtractListViewButtonInformation (_hToolbarWnd, _iButtonIndex);
	else if (UIOBJECT__TABCTRL_BUTTON == c_ObjectType)
		ExtractTabCtrlButtonInformation (_hToolbarWnd, _iButtonIndex);
	else if (UIOBJECT__LISTBOX_ITEM == c_ObjectType)
		ExtractListBoxItemInformation (_hToolbarWnd, _iButtonIndex);
	else if (UIOBJECT__COMBOBOX_ITEM == c_ObjectType)
		ExtractComboBoxItemInformation (_hToolbarWnd, _iButtonIndex);
	else if (UIOBJECT__TREEVIEW_ITEM == c_ObjectType)
		ASSERT (FALSE);

	s_Name.Replace ("&&", "\x1");
	s_Name.Replace ("&", "");
	s_Name.Replace ("\x1", "&");
	s_LowercaseName = s_Name;
	s_LowercaseName.LowerCase ();
	s_ID << (long)h_ParentWnd << '-' << (long)h_Wnd << '-' << c_ObjectType << '-' << i_CommandID;
	if (UIObject::p_CurrentSystemConfig->map_IdToUIObject.end () != UIObject::p_CurrentSystemConfig->map_IdToUIObject.find (s_ID))
	{
		s_ID = "";
		s_ID << (long)h_ParentWnd << '-' << (long)h_Wnd << '-' << c_ObjectType << '-' << i_CommandID << '-' << sl_ObjectCount;
	}
	++ sl_ObjectCount;
	b_IsVisible = _pParentWnd->b_IsVisible;
	b_IsSelfVisible = true;

	UIObject::set_ProcessIDs.insert (dw_ProcessID);
	UIObject::p_CurrentSystemConfig->dq_AllUIObjects.push_back (this);
	assert (UIObject::p_CurrentSystemConfig->map_IdToUIObject.end () == UIObject::p_CurrentSystemConfig->map_IdToUIObject.find (s_ID));
	UIObject::p_CurrentSystemConfig->map_IdToUIObject.insert (make_pair (s_ID, this));
	UIObject::p_CurrentSystemConfig->mmap_NameToUIObject.insert (make_pair (s_Name, this));
}


//
UIObject::UIObject (UIObject* _pParentObj, HWND _hTreeWnd, HTREEITEM _hItem, char _cObjectType)
{
	Init ();

	// toolbar button or listview item ...
	h_ParentWnd = _hTreeWnd;
	p_Parent = _pParentObj;
	ui_ShowState = _pParentObj->ui_ShowState;
	dw_ProcessID = _pParentObj->dw_ProcessID;

	c_ObjectType = _cObjectType;
	ExtractTreeViewItemInformation (_hTreeWnd, _hItem);

	s_Name.Replace ("&&", "\x1");
	s_Name.Replace ("&", "");
	s_Name.Replace ("\x1", "&");
	s_LowercaseName = s_Name;
	s_LowercaseName.LowerCase ();
	s_ID << (long)h_ParentWnd << '-' << (long)h_Wnd << '-' << c_ObjectType << '-' << i_CommandID;
	if (UIObject::p_CurrentSystemConfig->map_IdToUIObject.end () != UIObject::p_CurrentSystemConfig->map_IdToUIObject.find (s_ID))
	{
		s_ID = "";
		s_ID << (long)h_ParentWnd << '-' << (long)h_Wnd << '-' << c_ObjectType << '-' << i_CommandID << '-' << sl_ObjectCount;
	}
	++ sl_ObjectCount;
	b_IsVisible = _pParentObj->b_IsVisible;
	b_IsSelfVisible = true;

	UIObject::set_ProcessIDs.insert (dw_ProcessID);
	UIObject::p_CurrentSystemConfig->dq_AllUIObjects.push_back (this);
	assert (UIObject::p_CurrentSystemConfig->map_IdToUIObject.end () == UIObject::p_CurrentSystemConfig->map_IdToUIObject.find (s_ID));
	UIObject::p_CurrentSystemConfig->map_IdToUIObject.insert (make_pair (s_ID, this));
	UIObject::p_CurrentSystemConfig->mmap_NameToUIObject.insert (make_pair (s_Name, this));
}


//
UIObject::UIObject (UIObject* _pParentWnd, CMenu* _pMenu, int _iItemIndex)
{
	Init ();

	// menu item ...
	p_Parent = _pParentWnd;
	dw_ProcessID = _pParentWnd->dw_ProcessID;

	c_ObjectType = UIOBJECT__MENU_ITEM;
	ExtractMenuInformation (_pParentWnd, _pMenu, _iItemIndex);
	if (MF_SEPARATOR == i_MenuItemType)
		return;

	s_Name.Replace ("&&", "\x1");
	s_Name.Replace ("&", "");
	s_Name.Replace ("\x1", "&");
	s_LowercaseName = s_Name;
	s_LowercaseName.LowerCase ();
	s_ID << (long)h_ParentWnd << '-' << (long)h_Wnd << '-' << c_ObjectType << '-' << i_CommandID;
	if (UIObject::p_CurrentSystemConfig->map_IdToUIObject.end () != UIObject::p_CurrentSystemConfig->map_IdToUIObject.find (s_ID))
	{
		s_ID = "";
		s_ID << (long)h_ParentWnd << '-' << (long)h_Wnd << '-' << c_ObjectType << '-' << i_CommandID << '-' << sl_ObjectCount;
	}
	++ sl_ObjectCount;

	UIObject::set_ProcessIDs.insert (dw_ProcessID);
	//UIObject::p_CurrentSystemConfig->dq_AllUIObjects.push_back (this);
	assert (UIObject::p_CurrentSystemConfig->map_IdToUIObject.end () == UIObject::p_CurrentSystemConfig->map_IdToUIObject.find (s_ID));
	UIObject::p_CurrentSystemConfig->map_IdToUIObject.insert (make_pair (s_ID, this));
	UIObject::p_CurrentSystemConfig->mmap_NameToUIObject.insert (make_pair (s_Name, this));
}


//
UIObject::~UIObject (void)
{
	p_Parent = (UIObject*)0x1;
	p_MainWindow = (UIObject*)0x1;
	p_ActivationWindow = (UIObject*)0x1;
	h_Wnd = (HWND)0x1;
	h_ParentWnd = (HWND)0x1;

	IdToUIObject_map_t::iterator	iteMenu;
	for (iteMenu = map_CommandIdToMenuItems.begin ();
		iteMenu != map_CommandIdToMenuItems.end ();
		++ iteMenu)
		delete iteMenu->second;

	IdToUIObject_map_t::iterator	ite;
	for (ite = map_IdToChildren.begin ();
		ite != map_IdToChildren.end ();
		++ ite)
		delete ite->second;
}


//
bool UIObject::ExtractWindowInformation (CWnd* _pWnd)
{
	//
	h_Wnd = _pWnd->GetSafeHwnd ();

	//
	CString sName;
	int nLen = ::GetWindowTextLength(h_Wnd);
	::GetWindowText(h_Wnd, sName.GetBufferSetLength(nLen), nLen+1);
	sName.ReleaseBuffer();
	s_Name = (LPCSTR) sName;
	i_CommandID = -1;

	//
	char zClassName [1001];
	zClassName [0] = '\0';
	GetClassName (h_Wnd, zClassName, 1000);
	s_ClassName = zClassName;
	if ("#32769" == s_ClassName)
		s_ClassName = "Desktop";
	else if ("#32768" == s_ClassName)
	{
		c_ObjectType = UIOBJECT__POPUP_MENU;
		s_ClassName = "Popup Menu";
	}
	else if ("#32770" == s_ClassName)
	{
		c_ObjectType = UIOBJECT__DIALOG;
		s_ClassName = "Dialog Box";
	}
	else if ("DV2ControlHost" == s_ClassName)
	{
		b_StartMenu = true;
		UIObject::b_StartMenuIdentified = true;
	}

	WindowClassToCount_map_t::iterator iteCount;
	iteCount = UIObject::map_WindowClassToCount.find (s_ClassName);
	if (iteCount == UIObject::map_WindowClassToCount.end ())
		UIObject::map_WindowClassToCount.insert (make_pair (s_ClassName, 1));
	else
		++ iteCount->second;

	// the "Start" button on Win2000 doesn't seem to have text on it, so we need to set the name specifically...
	if (("" == s_Name) && ("Button" == s_ClassName))
	{
		CWnd* pParent = _pWnd->GetParent ();
		char zParentClass [1001];
		GetClassName (pParent->GetSafeHwnd (), zParentClass, 1000);
		if (0 == strcmp (zParentClass, "Shell_TrayWnd"))
		{
			s_Name = "start";
			b_StartButtonIdentified = true;
		}
	}

	//
	h_ParentWnd = ::GetParent (h_Wnd);

	//
	rct_Body.SetRectEmpty ();
	::GetWindowRect (h_Wnd, &rct_Body);

	//
	b_IsVisible = (TRUE == ::IsWindowVisible (h_Wnd));
	b_IsSelfVisible = b_IsVisible;

	//
	HWND hFocusWindow = ::GetFocus ();
	if (NULL == hFocusWindow)
		b_HasFocus = false;
	else
		b_HasFocus = (hFocusWindow == h_Wnd);

	//
	HWND hActiveWindow = ::GetActiveWindow ();
	if (NULL == hActiveWindow)
		b_IsActive = false;
	else
		b_IsActive = (hActiveWindow == h_Wnd);

	//
	HWND hForegroundWnd = ::GetForegroundWindow ();
	if (NULL == hForegroundWnd)
		b_IsForeground = false;
	else
		b_IsForeground = (hForegroundWnd == h_Wnd);

	//
	b_IsEnabledForInput = (TRUE == ::IsWindowEnabled (h_Wnd));

	//
	dw_Style = (DWORD)GetWindowLong (h_Wnd, GWL_STYLE);
	if ("Button" == s_ClassName)
	{
		DWORD dwButtonType = (BS_TYPEMASK & dw_Style);
		if ((BS_RADIOBUTTON == dwButtonType) || (BS_AUTORADIOBUTTON == dwButtonType))
			s_ClassName = "radiobutton";
		else if ((BS_CHECKBOX == dwButtonType) ||
				 (BS_AUTOCHECKBOX == dwButtonType) ||
				 (BS_AUTO3STATE == dwButtonType) ||
				 (BS_3STATE == dwButtonType))
			s_ClassName = "checkbox";
		if (BS_GROUPBOX == dwButtonType)
			s_ClassName = "groupbox";
		if (WS_DISABLED & dw_Style)
		{
			b_IsVisible = false;
			b_IsSelfVisible = false;
		}
	}
	//
	dw_ExStyle = (DWORD)GetWindowLong (h_Wnd, GWL_EXSTYLE);

	//
	b_IsChildWnd = (0 != (WS_CHILD & dw_Style));

	//
	WINDOWPLACEMENT wp;
	wp.length = sizeof(WINDOWPLACEMENT);
	::GetWindowPlacement(h_Wnd, &wp);
	ui_ShowState = wp.showCmd;

	//
	if ("ToolbarWindow32" == s_ClassName)
		ExtractToolbarInformation (this);
	else if ("SysListView32" == s_ClassName)
		ExtractListViewInformation (this);
	else if ("SysTabControl32" == s_ClassName)
		ExtractTabControlInformation (this);
	else if ("ListBox" == s_ClassName)
		ExtractListBoxInformation (this);
	else if ("ComboBox" == s_ClassName)
		ExtractComboBoxInformation (this);
	else if ("SysTreeView32" == s_ClassName)
		ExtractTreeViewInformation ();
	if (UIOBJECT__POPUP_MENU == c_ObjectType)
	{
		HMENU hMenu = (HMENU) SendMessage (h_Wnd, MN_GETHMENU, 0, 0);
		CMenu* pMenu = NULL;
		if (NULL != hMenu)
			pMenu = CMenu::FromHandle (hMenu);
		if (NULL != pMenu)
		{
			UINT iItemCount = pMenu->GetMenuItemCount ();
			for (int i = 0; i < (int)iItemCount; ++ i)
			{
				UIObject* pObject = new UIObject (this, pMenu, i);
				if (MF_SEPARATOR == pObject->i_MenuItemType)
				{
					delete pObject;
					continue;
				}
				pObject->p_Parent = this;
				pObject->b_IsVisible = b_IsVisible;
				pObject->b_IsSelfVisible = true;
				pObject->b_HasFocus = b_HasFocus;
				pObject->b_IsActive = b_IsActive;
				pObject->b_IsForeground = b_IsForeground;
				pObject->b_IsEnabledForInput = b_IsEnabledForInput;
				pObject->ui_ShowState = ui_ShowState;

				map_CommandIdToMenuItems.insert (make_pair (pObject->s_ID, pObject));
			}
		}
	}

	//
	if ((false == b_IsChildWnd) && (TRUE == ::IsWindow (h_Wnd)))
	{
		if (FALSE == ::IsWindow (h_Wnd))
			return false;
		HMENU hMenu = ::GetMenu(h_Wnd);
		CMenu* pMenu = NULL;
		if (NULL != hMenu)
			pMenu = CMenu::FromHandle(hMenu);
		if (NULL != pMenu)
		{
			h_Menu = pMenu->m_hMenu;
			UINT iItemCount = pMenu->GetMenuItemCount ();
			for (int i = 0; i < (int)iItemCount; ++ i)
			{
				UIObject* pObject = new UIObject (this, pMenu, i);
				if (MF_SEPARATOR == pObject->i_MenuItemType)
				{
					delete pObject;
					continue;
				}
				pObject->p_Parent = this;
				pObject->b_IsVisible = b_IsVisible;
				pObject->b_IsSelfVisible = true;
				pObject->b_HasFocus = b_HasFocus;
				pObject->b_IsActive = b_IsActive;
				pObject->b_IsForeground = b_IsForeground;
				pObject->b_IsEnabledForInput = b_IsEnabledForInput;
				pObject->ui_ShowState = ui_ShowState;

				map_CommandIdToMenuItems.insert (make_pair (pObject->s_ID, pObject));
			}
		}
	}
	if (FALSE == ::IsWindow (h_Wnd))
		return false;
	return true;
}


//
void UIObject::ExtractToolbarInformation (UIObject* _pUIObject)
{
	long lButtonCount = (long) SendMessage (h_Wnd, TB_BUTTONCOUNT, 0, 0);
	for (long i = 0; i < lButtonCount; ++ i)
		UIObject* pObject = new UIObject (_pUIObject, h_Wnd, i, UIOBJECT__TOOLBAR_BUTTON);
}


//
void UIObject::ExtractToolbarButtonInformation (HWND _hToolbarWnd, int _iButtonIndex)
{
	//
	h_Wnd = NULL;
	h_ParentWnd = _hToolbarWnd;
	h_Menu = NULL;

	//
	s_ClassName = "ToolbarButton";

	// --------------------------------------------------------------------------------------
	// the rest of the toolbar button information is a little more complicated to access ...
	// --------------------------------------------------------------------------------------
	// get access to the process that created the window we're working with ...
	DWORD iProcessID = 0;
	GetWindowThreadProcessId (h_ParentWnd, &iProcessID);
	HANDLE hProcess = OpenProcess (PROCESS_VM_READ|PROCESS_VM_OPERATION, 0, iProcessID);
	if (NULL == hProcess)
		return;

	// allocated memory accessible by the process we retrieved above ...
	LPVOID lpRemoteRect = VirtualAllocEx (hProcess, NULL, sizeof (RECT), MEM_COMMIT, PAGE_READWRITE);
	LPVOID lpRemoteButtonInfo = VirtualAllocEx (hProcess, NULL, sizeof (TBBUTTON), MEM_COMMIT, PAGE_READWRITE);
	LPVOID lpRemoteText = VirtualAllocEx (hProcess, NULL, sizeof (char) * 10001, MEM_COMMIT, PAGE_READWRITE);

	//
	if (TRUE == SendMessage (h_ParentWnd, TB_GETITEMRECT, (WPARAM)_iButtonIndex, (LPARAM)lpRemoteRect))
	{
		RECT rect;
		DWORD iBytesRead;
		ReadProcessMemory (hProcess, lpRemoteRect, &rect, sizeof(RECT), &iBytesRead);
		rct_Body = rect;
	}
	else
		rct_Body.SetRectEmpty ();

	//
	if (TRUE == SendMessage (h_ParentWnd, TB_GETBUTTON, (WPARAM)_iButtonIndex, (LPARAM)lpRemoteButtonInfo))
	{
		TBBUTTON ButtonInfo;
		DWORD iBytesRead;
		ReadProcessMemory (hProcess, lpRemoteButtonInfo, &ButtonInfo, sizeof(TBBUTTON), &iBytesRead);
		i_CommandID = ButtonInfo.idCommand;
	}
	//
	long lLength = SendMessage (h_ParentWnd, TB_GETBUTTONTEXT, (WPARAM)i_CommandID, (LPARAM)lpRemoteText);
	if (-1 != lLength)
	{
		char zName [10000];
		DWORD iBytesRead;
		ReadProcessMemory (hProcess, lpRemoteText, zName, sizeof(char) * lLength, &iBytesRead);
		zName [lLength] = '\0';
		s_Name = (char*)zName;
	}
	//
	/*
	if ("" == s_Name)
	{
		HWND hTooltip = (HWND)SendMessage (h_ParentWnd, TB_GETTOOLTIPS, 0, 0);
		if (NULL != hTooltip)
		{
			TOOLINFO ti;
			memset (&ti, 0, sizeof (TOOLINFO));
			ti.cbSize = sizeof (TOOLINFO);
			ti.hwnd = hTooltip;
			ti.hinst = NULL;
			ti.lpszText = (LPSTR) lpRemoteText;
			ti.uId = i_CommandID;

			LPVOID lpRemoteTip = VirtualAllocEx (hProcess, NULL, sizeof (TOOLINFO), MEM_COMMIT, PAGE_READWRITE);
			DWORD iBytesWritten;
			WriteProcessMemory (hProcess, lpRemoteTip, &ti, sizeof (TOOLINFO), &iBytesWritten);

			SendMessage (hTooltip, TTM_GETTEXT, (WPARAM)10000, (LPARAM)lpRemoteTip);
			DWORD iBytesRead;
			ReadProcessMemory (hProcess, lpRemoteTip, &ti, sizeof(TOOLINFO), &iBytesRead);

			char zName [10001];
			ReadProcessMemory (hProcess, lpRemoteText, zName, sizeof(char) * 10000, &iBytesRead);
			s_Name = zName;

			VirtualFreeEx (hProcess, lpRemoteTip, 0, MEM_RELEASE);
		}
	}*/

	VirtualFreeEx (hProcess, lpRemoteRect, 0, MEM_RELEASE);
	VirtualFreeEx (hProcess, lpRemoteButtonInfo, 0, MEM_RELEASE);
	VirtualFreeEx (hProcess, lpRemoteText, 0, MEM_RELEASE);
}


//
void UIObject::ExtractListViewInformation (UIObject* _pUIObject)
{
	long lButtonCount = (long) SendMessage (h_Wnd, LVM_GETITEMCOUNT, 0, 0);
	for (long i = 0; i < lButtonCount; ++ i)
		new UIObject (_pUIObject, h_Wnd, i, UIOBJECT__LISTVIEW_BUTTON);
}


//
void UIObject::ExtractListViewButtonInformation (HWND _hListViewWnd, int _iItemIndex)
{
	//
	h_Wnd = NULL;
	h_ParentWnd = _hListViewWnd;
	h_Menu = NULL;
	i_CommandID = _iItemIndex;

	//
	s_ClassName = "ListViewButton";

	// --------------------------------------------------------------------------------------
	// the rest of the toolbar button information is a little more complicated to access ...
	// --------------------------------------------------------------------------------------
	// get access to the process that created the window we're working with ...
	DWORD iProcessID = 0;
	GetWindowThreadProcessId (h_ParentWnd, &iProcessID);
	HANDLE hProcess = OpenProcess (PROCESS_VM_READ|PROCESS_VM_WRITE|PROCESS_VM_OPERATION, 0, iProcessID);
	if (NULL == hProcess)
		return;

	// allocated memory accessible by the process we retrieved above ...
	LPVOID lpRemoteRect = VirtualAllocEx (hProcess, NULL, sizeof (RECT), MEM_COMMIT, PAGE_READWRITE);
	LPVOID lpRemoteItemInfo = VirtualAllocEx (hProcess, NULL, sizeof (LVITEM), MEM_COMMIT, PAGE_READWRITE);
	LPVOID lpRemoteText = VirtualAllocEx (hProcess, NULL, sizeof (char) * 10000, MEM_COMMIT, PAGE_READWRITE);

	//
	if (TRUE == SendMessage (h_ParentWnd, LVM_GETITEMRECT, (WPARAM)_iItemIndex, (LPARAM)lpRemoteRect))
	{
		RECT rect;
		DWORD iBytesRead;
		ReadProcessMemory (hProcess, lpRemoteRect, &rect, sizeof(RECT), &iBytesRead);
		rct_Body = rect;
	}
	else
		rct_Body.SetRectEmpty ();

	// get item text ...
	{
		LVITEM oItemInfo;
		oItemInfo.mask = LVIF_TEXT | LVIF_STATE;
		oItemInfo.iItem = _iItemIndex;
		oItemInfo.iSubItem = 0;
		oItemInfo.pszText = (LPSTR)lpRemoteText;
		oItemInfo.cchTextMax = 10000;

		DWORD iBytesWritten;
		WriteProcessMemory (hProcess, lpRemoteItemInfo, &oItemInfo, sizeof (LVITEM), &iBytesWritten);
		long lLength = SendMessage (h_ParentWnd, LVM_GETITEMTEXT, (WPARAM)_iItemIndex, (LPARAM)lpRemoteItemInfo);
		if (-1 != lLength)
		{
			DWORD iBytesRead;
			char zName [10000];
			ReadProcessMemory (hProcess, lpRemoteText, zName, sizeof(char) * lLength, &iBytesRead);
			zName [lLength] = '\0';
			s_Name = (char*)zName;
		}
	}
	// get first subitem text ...
	{
		LVITEM oItemInfo;
		oItemInfo.mask = LVIF_TEXT;
		oItemInfo.iItem = _iItemIndex;
		oItemInfo.iSubItem = 1;
		oItemInfo.pszText = (LPSTR)lpRemoteText;
		oItemInfo.cchTextMax = 10000;

		DWORD iBytesWritten;
		WriteProcessMemory (hProcess, lpRemoteItemInfo, &oItemInfo, sizeof (LVITEM), &iBytesWritten);
		long lLength = SendMessage (h_ParentWnd, LVM_GETITEMTEXT, (WPARAM)_iItemIndex, (LPARAM)lpRemoteItemInfo);
		if (-1 != lLength)
		{
			DWORD iBytesRead;
			char zName [10000];
			ReadProcessMemory (hProcess, lpRemoteText, zName, sizeof(char) * lLength, &iBytesRead);
			zName [lLength] = '\0';
			if ((strlen (zName) > 0) && (s_Name == ""))
				s_Name = (char*)zName;

			s_ExtendedName = "";
			if (s_Name != zName)
			{
				s_ExtendedName << s_Name << " " << zName;
				s_ExtendedName.Strip ();
				s_ExtendedName.LowerCase ();
			}
		}
	}

	VirtualFreeEx (hProcess, lpRemoteRect, 0, MEM_RELEASE);
	VirtualFreeEx (hProcess, lpRemoteItemInfo, 0, MEM_RELEASE);
	VirtualFreeEx (hProcess, lpRemoteText, 0, MEM_RELEASE);
}



//
void UIObject::ExtractTabControlInformation (UIObject* _pUIObject)
{
	long lTabCount = (long) SendMessage (h_Wnd, TCM_GETITEMCOUNT, 0, 0);
	for (long i = 0; i < lTabCount; ++ i)
		UIObject* pObject = new UIObject (_pUIObject, h_Wnd, i, UIOBJECT__TABCTRL_BUTTON);
}


//
void UIObject::ExtractTabCtrlButtonInformation (HWND _hTabCtrlWnd, int _iTabIndex)
{
	//
	h_Wnd = NULL;
	h_ParentWnd = _hTabCtrlWnd;
	h_Menu = NULL;
	i_CommandID = _iTabIndex;

	//
	s_ClassName = "TabCtrlButton";

	// --------------------------------------------------------------------------------------
	// the rest of the toolbar button information is a little more complicated to access ...
	// --------------------------------------------------------------------------------------
	// get access to the process that created the window we're working with ...
	DWORD iProcessID = 0;
	GetWindowThreadProcessId (h_ParentWnd, &iProcessID);
	HANDLE hProcess = OpenProcess (PROCESS_VM_READ|PROCESS_VM_WRITE|PROCESS_VM_OPERATION, 0, iProcessID);
	if (NULL == hProcess)
		return;

	// allocated memory accessible by the process we retrieved above ...
	LPVOID lpRemoteRect = VirtualAllocEx (hProcess, NULL, sizeof (RECT), MEM_COMMIT, PAGE_READWRITE);
	LPVOID lpRemoteTabInfo = VirtualAllocEx (hProcess, NULL, sizeof (TCITEM), MEM_COMMIT, PAGE_READWRITE);
	LPVOID lpRemoteText = VirtualAllocEx (hProcess, NULL, sizeof (char) * 10000, MEM_COMMIT, PAGE_READWRITE);

	//
	if (TRUE == SendMessage (h_ParentWnd, TCM_GETITEMRECT, (WPARAM)_iTabIndex, (LPARAM)lpRemoteRect))
	{
		RECT rect;
		DWORD iBytesRead;
		ReadProcessMemory (hProcess, lpRemoteRect, &rect, sizeof(RECT), &iBytesRead);
		rct_Body = rect;
	}
	else
		rct_Body.SetRectEmpty ();


	TCITEM oTabInfo;
	oTabInfo.mask = TCIF_TEXT|TCIF_STATE;
	oTabInfo.pszText = (LPSTR) lpRemoteText;
	oTabInfo.cchTextMax = 10000;

	DWORD iBytesWritten;
	WriteProcessMemory (hProcess, lpRemoteTabInfo, &oTabInfo, sizeof (LVITEM), &iBytesWritten);
	//
	if (TRUE == SendMessage (h_ParentWnd, TCM_GETITEM, (WPARAM)_iTabIndex, (LPARAM)lpRemoteTabInfo))
	{
		DWORD iBytesRead;
		ReadProcessMemory (hProcess, lpRemoteTabInfo, &oTabInfo, sizeof(TCITEM), &iBytesRead);

		if (NULL != oTabInfo.pszText)
		{
			char zName [10000];
			ReadProcessMemory (hProcess, lpRemoteText, zName, sizeof(char) * 10000, &iBytesRead);
			s_Name = (char*)zName;
		}
	}

	VirtualFreeEx (hProcess, lpRemoteRect, 0, MEM_RELEASE);
	VirtualFreeEx (hProcess, lpRemoteTabInfo, 0, MEM_RELEASE);
	VirtualFreeEx (hProcess, lpRemoteText, 0, MEM_RELEASE);
}


//
void UIObject::ExtractListBoxInformation (UIObject* _pUIObject)
{
	long lItemCount = (long) SendMessage (h_Wnd, LB_GETCOUNT, 0, 0);
	for (long i = 0; i < lItemCount; ++ i)
		UIObject* pObject = new UIObject (_pUIObject, h_Wnd, i, UIOBJECT__LISTBOX_ITEM);
}


//
void UIObject::ExtractListBoxItemInformation (HWND _hListBoxWnd, int _iItemIndex)
{
	//
	h_Wnd = NULL;
	h_ParentWnd = _hListBoxWnd;
	h_Menu = NULL;
	i_CommandID = _iItemIndex;

	//
	s_ClassName = "ListBoxItem";

	// --------------------------------------------------------------------------------------
	// the rest of the toolbar button information is a little more complicated to access ...
	// --------------------------------------------------------------------------------------
	// get access to the process that created the window we're working with ...
	DWORD iProcessID = 0;
	GetWindowThreadProcessId (h_ParentWnd, &iProcessID);
	HANDLE hProcess = OpenProcess (PROCESS_VM_READ|PROCESS_VM_WRITE|PROCESS_VM_OPERATION, 0, iProcessID);
	if (NULL == hProcess)
		return;

	// allocated memory accessible by the process we retrieved above ...
	LPVOID lpRemoteRect = VirtualAllocEx (hProcess, NULL, sizeof (RECT), MEM_COMMIT, PAGE_READWRITE);
	LPVOID lpRemoteText = VirtualAllocEx (hProcess, NULL, sizeof (char) * 10000, MEM_COMMIT, PAGE_READWRITE);

	//
	if (TRUE == SendMessage (h_ParentWnd, LB_GETITEMRECT, (WPARAM)_iItemIndex, (LPARAM)lpRemoteRect))
	{
		RECT rect;
		DWORD iBytesRead;
		ReadProcessMemory (hProcess, lpRemoteRect, &rect, sizeof(RECT), &iBytesRead);
		rct_Body = rect;
	}
	else
		rct_Body.SetRectEmpty ();

	//
	long lLength = SendMessage (h_ParentWnd, LB_GETTEXT, (WPARAM)_iItemIndex, (LPARAM)lpRemoteText);
	if (-1 != lLength)
	{
		DWORD iBytesRead;
		char zName [10000];
		ReadProcessMemory (hProcess, lpRemoteText, zName, sizeof(char) * lLength, &iBytesRead);
		zName [lLength] = '\0';
		s_Name = (char*)zName;
	}
	else
	{
		char zName [10000];
		lLength = SendMessage (h_ParentWnd, LB_GETTEXT, (WPARAM)_iItemIndex, (LPARAM)zName);
		s_Name = zName;
	}

	VirtualFreeEx (hProcess, lpRemoteRect, 0, MEM_RELEASE);
	VirtualFreeEx (hProcess, lpRemoteText, 0, MEM_RELEASE);
}


//
void UIObject::ExtractComboBoxInformation (UIObject* _pUIObject)
{
	long lItemCount = (long) SendMessage (h_Wnd, CB_GETCOUNT, 0, 0);
	for (long i = 0; i < lItemCount; ++ i)
		UIObject* pObject = new UIObject (_pUIObject, h_Wnd, i, UIOBJECT__COMBOBOX_ITEM);
}


//
void UIObject::ExtractComboBoxItemInformation (HWND _hComboBoxWnd, int _iItemIndex)
{
	//
	h_Wnd = NULL;
	h_ParentWnd = _hComboBoxWnd;
	h_Menu = NULL;
	i_CommandID = _iItemIndex;

	//
	s_ClassName = "ComboBoxItem";

	// GetDroppedState

	// --------------------------------------------------------------------------------------
	// the rest of the toolbar button information is a little more complicated to access ...
	// --------------------------------------------------------------------------------------
	// get access to the process that created the window we're working with ...
	DWORD iProcessID = 0;
	GetWindowThreadProcessId (h_ParentWnd, &iProcessID);
	HANDLE hProcess = OpenProcess (PROCESS_VM_READ|PROCESS_VM_WRITE|PROCESS_VM_OPERATION, 0, iProcessID);
	if (NULL == hProcess)
		return;

	//
	long lLength = SendMessage (h_ParentWnd, CB_GETLBTEXTLEN, (WPARAM)_iItemIndex, 0);
	char* zName = new char [lLength + 10];
	lLength = SendMessage (h_ParentWnd, CB_GETLBTEXT, (WPARAM)_iItemIndex, (LPARAM)zName);
	zName [lLength] = '\0';
	s_Name = zName;
	delete[] zName;

	/*if ("" == s_Name)
	{
		LPVOID lpRemoteText = VirtualAllocEx (hProcess, NULL, sizeof (char) * 10000, MEM_COMMIT, PAGE_READWRITE);
		long lLength = SendMessage (h_ParentWnd, CB_GETLBTEXT, (WPARAM)_iItemIndex, (LPARAM)lpRemoteText);
		if (-1 != lLength)
		{
			DWORD iBytesRead;
			char zName [10000];
			ReadProcessMemory (hProcess, lpRemoteText, zName, sizeof(char) * lLength, &iBytesRead);
			zName [lLength] = '\0';
			s_Name = (char*)zName;
		}
		VirtualFreeEx (hProcess, lpRemoteText, 0, MEM_RELEASE);
	}*/
}


//
void UIObject::ExtractTreeViewInformation (void)
{
	HTREEITEM hItem = (HTREEITEM) SendMessage (h_Wnd, TVM_GETNEXTITEM, TVGN_ROOT, 0);
	while (NULL != hItem)
	{
		UIObject* pObject = new UIObject (this, h_Wnd, hItem, UIOBJECT__TREEVIEW_ITEM);
		pObject->p_Parent = this;
		AddChild (pObject);
		pObject->ExtractTreeViewChildren (h_Wnd, hItem);

		hItem = (HTREEITEM) SendMessage (h_Wnd, TVM_GETNEXTITEM, TVGN_NEXT, (LPARAM)hItem);
	}
}


//
void UIObject::ExtractTreeViewChildren (HWND _hTreeWnd, HTREEITEM _hItem)
{
	HTREEITEM hItem = (HTREEITEM) SendMessage (_hTreeWnd, TVM_GETNEXTITEM, TVGN_CHILD, (LPARAM)_hItem);
	while (NULL != hItem)
	{
		UIObject* pObject = new UIObject (this, _hTreeWnd, hItem, UIOBJECT__TREEVIEW_ITEM);
		pObject->p_Parent = this;
		AddChild (pObject);
		pObject->ExtractTreeViewChildren (_hTreeWnd, hItem);

		hItem = (HTREEITEM) SendMessage (_hTreeWnd, TVM_GETNEXTITEM, TVGN_NEXT, (LPARAM)hItem);
	}
}


//
void UIObject::ExtractTreeViewItemInformation (HWND _hTreeView, HTREEITEM _hItem)
{
	//
	h_Wnd = NULL;
	h_ParentWnd = _hTreeView;
	h_Menu = NULL;
	i_CommandID = (long)_hItem;

	//
	s_ClassName = "TreeViewButton";

	// --------------------------------------------------------------------------------------
	// the rest of the toolbar button information is a little more complicated to access ...
	// --------------------------------------------------------------------------------------
	// get access to the process that created the window we're working with ...
	DWORD iProcessID = 0;
	GetWindowThreadProcessId (h_ParentWnd, &iProcessID);
	HANDLE hProcess = OpenProcess (PROCESS_VM_READ|PROCESS_VM_WRITE|PROCESS_VM_OPERATION, 0, iProcessID);
	if (NULL == hProcess)
		return;

	// allocated memory accessible by the process we retrieved above ...
	LPVOID lpRemoteRect = VirtualAllocEx (hProcess, NULL, sizeof (RECT), MEM_COMMIT, PAGE_READWRITE);
	LPVOID lpRemoteItemInfo = VirtualAllocEx (hProcess, NULL, sizeof (TVITEM), MEM_COMMIT, PAGE_READWRITE);
	LPVOID lpRemoteText = VirtualAllocEx (hProcess, NULL, sizeof (char) * 10001, MEM_COMMIT, PAGE_READWRITE);

	//
	TVITEM oItemInfo;
	oItemInfo.mask = TVIF_TEXT | TVIF_HANDLE;
	oItemInfo.hItem = _hItem;
	oItemInfo.pszText = (LPSTR)lpRemoteText;
	oItemInfo.cchTextMax = 10000;

	DWORD iBytesWritten;
	WriteProcessMemory (hProcess, lpRemoteItemInfo, &oItemInfo, sizeof (TVITEM), &iBytesWritten);
	if (TRUE == SendMessage (h_ParentWnd, TVM_GETITEM, (WPARAM)0, (LPARAM)lpRemoteItemInfo))
	{
		DWORD iBytesRead;
		ReadProcessMemory (hProcess, lpRemoteItemInfo, &oItemInfo, sizeof(TVITEM), &iBytesRead);

		char zName [10001];
		ReadProcessMemory (hProcess, lpRemoteText, zName, sizeof(char) * oItemInfo.cchTextMax, &iBytesRead);
		zName [oItemInfo.cchTextMax] = '\0';
		s_Name = (char*)zName;
	}

	//
	RECT rect;
	*(HTREEITEM*) &rect = _hItem;
	WriteProcessMemory (hProcess, lpRemoteRect, &rect, sizeof (RECT), &iBytesWritten);
	if (TRUE == SendMessage (h_ParentWnd, TVM_GETITEMRECT, (WPARAM)TRUE, (LPARAM)lpRemoteRect))
	{
		RECT rect;
		DWORD iBytesRead;
		ReadProcessMemory (hProcess, lpRemoteRect, &rect, sizeof(RECT), &iBytesRead);
		rct_Body = rect;
	}
	else
		rct_Body.SetRectEmpty ();


	VirtualFreeEx (hProcess, lpRemoteRect, 0, MEM_RELEASE);
	VirtualFreeEx (hProcess, lpRemoteItemInfo, 0, MEM_RELEASE);
	VirtualFreeEx (hProcess, lpRemoteText, 0, MEM_RELEASE);
}


//
void UIObject::ExtractMenuInformation (UIObject* _pParentWnd, CMenu* _pMenu, int _iItemIndex)
{
	MENUITEMINFO mii;

	mii.cbSize = sizeof (MENUITEMINFO);
	mii.fMask = MIIM_CHECKMARKS | MIIM_FTYPE | MIIM_ID | MIIM_STATE | MIIM_STRING | MIIM_SUBMENU;
	mii.dwTypeData = NULL;
	_pMenu->GetMenuItemInfoA (_iItemIndex, &mii, TRUE);

	h_ParentWnd = _pParentWnd->h_Wnd;
	s_ClassName = "Menu";
	i_CommandID = mii.wID;
	i_MenuItemType = mii.fType;
	i_MenuItemState = mii.fState;

	HMENU hSubMenu = mii.hSubMenu;
	String sName;
	//if (MFT_STRING == mii.fType)
	{
		char* pzName = new char [mii.cch + 1];
		mii.dwTypeData = (LPTSTR) pzName;
		++ mii.cch;
		pzName [0] = '\0';
		_pMenu->GetMenuItemInfoA (_iItemIndex, &mii, TRUE);
		s_Name = pzName;
		delete[] pzName;
	}
	GetMenuItemRect (h_ParentWnd, _pMenu->m_hMenu, _iItemIndex, &rct_Body);


	if (NULL != hSubMenu)
	{
		CMenu* pSubMenu = CMenu::FromHandle (hSubMenu);
		if (NULL == pSubMenu->GetSafeHmenu ())
			return;
		UINT iItemCount = pSubMenu->GetMenuItemCount ();
		for (int i = 0; i < (int)iItemCount; ++ i)
		{
			UIObject* pObject = new UIObject (_pParentWnd, pSubMenu, i);
			pObject->p_Parent = this;
			if (MF_SEPARATOR == pObject->i_MenuItemType)
			{
				delete pObject;
				continue;
			}
			assert (map_IdToChildren.end () == map_IdToChildren.find (pObject->s_ID));
			map_IdToChildren.insert (make_pair (pObject->s_ID, pObject));
		}
	}
}


//
void UIObject::AddChild (UIObject* _pUIObject)
{
	map_IdToChildren.insert (make_pair (_pUIObject->s_ID, _pUIObject));
}


//
void UIObject::ScanFromWindow (CWnd* _pWnd, ProcessID_set_t* _psetProcessesToIgnore)
{
	if (CWnd::GetDesktopWindow () == _pWnd)
	{
		UIObject::sl_ObjectCount = 0;
		if (NULL == _psetProcessesToIgnore)
		{
			UIObject::b_StartButtonIdentified = false;
			UIObject::b_StartMenuIdentified = false;
			UIObject::map_WindowClassToCount.clear ();
		}
	}

	CWnd* pWnd = _pWnd;
	while (NULL != pWnd)
	{
		HWND hWnd = pWnd->GetSafeHwnd ();
		if (FALSE == ::IsWindow(hWnd))
		{
			HWND hNextWnd = ::GetNextWindow (hWnd, GW_HWNDNEXT);
			pWnd = (NULL == hNextWnd)? NULL : CWnd::FromHandle (hNextWnd);
			continue;
		}

		// check and ignore my own windows and any windows in the process ignore list ...
		DWORD dwProcessID = 0;
		GetWindowThreadProcessId (pWnd->GetSafeHwnd (), &dwProcessID);
		bool bSkip = false;
		if (dwProcessID == UIObject::dw_MyProcessID)
			bSkip = true;
		else if (NULL != _psetProcessesToIgnore)
		{
			if (_psetProcessesToIgnore->end () == _psetProcessesToIgnore->find (dwProcessID))
				bSkip = true;
		}
		if (true == bSkip)
		{
			HWND hNextWnd = ::GetNextWindow (hWnd, GW_HWNDNEXT);
			pWnd = (NULL == hNextWnd)? NULL : CWnd::FromHandle (hNextWnd);
			continue;
		}


		// ignore any windows that are in the window class or window ignore list ...
		CString sName;
		char zClassName [1001];
		int nLen = ::GetWindowTextLength(hWnd);
		::GetWindowText(hWnd, sName.GetBufferSetLength(nLen), nLen+1);
		sName.ReleaseBuffer();
		GetClassName (hWnd, zClassName, 1000);

		if ((UIObject::set_WindowsToIgnore.end () != UIObject::set_WindowsToIgnore.find ((LPCSTR) sName)) ||
			(UIObject::set_ClassesToIgnore.end () != UIObject::set_ClassesToIgnore.find (zClassName)))
		{
			HWND hNextWnd = ::GetNextWindow (hWnd, GW_HWNDNEXT);
			pWnd = (NULL == hNextWnd)? NULL : CWnd::FromHandle (hNextWnd);
			continue;
		}
		HWND hParentWnd = hWnd;
		while (NULL != GetParent (hParentWnd))
		{
			HWND hNextWnd = GetParent (hParentWnd);
			if (TRUE == ::IsWindow (hNextWnd))
				hParentWnd = hNextWnd;
			else
				break;
		}

		CWnd* pParentWnd = CWnd::FromHandle (hParentWnd);
		sName = "";
		nLen = ::GetWindowTextLength(hParentWnd);
		::GetWindowText(hParentWnd, sName.GetBufferSetLength(nLen), nLen+1);
		sName.ReleaseBuffer();
		GetClassName (hParentWnd, zClassName, 1000);
		if ((UIObject::set_WindowsToIgnore.end () != UIObject::set_WindowsToIgnore.find ((LPCSTR) sName)) ||
			(UIObject::set_ClassesToIgnore.end () != UIObject::set_ClassesToIgnore.find (zClassName)))
		{
			HWND hNextWnd = ::GetNextWindow (hWnd, GW_HWNDNEXT);
			pWnd = (NULL == hNextWnd)? NULL : CWnd::FromHandle (hNextWnd);
			continue;
		}


		//
		if (FALSE == ::IsWindow(hWnd))
		{
			HWND hNextWnd = ::GetNextWindow (hWnd, GW_HWNDNEXT);
			pWnd = (NULL == hNextWnd)? NULL : CWnd::FromHandle (hNextWnd);
			continue;
		}

		new UIObject (pWnd);
		HWND hChild = GetTopWindow (hWnd);
		CWnd* pChild = (NULL == hChild)? NULL : CWnd::FromHandle (hChild);
		if (NULL != pChild)
			UIObject::ScanFromWindow (pChild, _psetProcessesToIgnore);

		HWND hNextWnd = ::GetNextWindow (hWnd, GW_HWNDNEXT);
		pWnd = (NULL == hNextWnd)? NULL : CWnd::FromHandle (hNextWnd);
	}
}



//
void UIObject::Clear (void)
{
	UIObject::map_WindowClassToCount.clear ();
}


//
void UIObject::ClearTemporaryObjects (void)
{
	UIObject_dq_t::iterator	ite;
	for (ite = UIObject::sdq_TemporaryObjects.begin (); 
		ite != UIObject::sdq_TemporaryObjects.end ();
		++ ite)
		delete *ite;
	UIObject::sdq_TemporaryObjects.clear ();
}


//
void UIObject::LoadIgnoreList (void)
{
	File::ReadLines ("classes_to_ignore.txt", set_ClassesToIgnore);
	File::ReadLines ("windows_to_ignore.txt", set_WindowsToIgnore);
}


//
UIObject* UIObject::Clone (UIObject* _pParent)
{
	UIObject* pClone = new UIObject;

	pClone->c_ObjectType = c_ObjectType;
	pClone->s_Name = s_Name;
	pClone->s_LowercaseName = s_LowercaseName;
	pClone->s_ClassName = s_ClassName;
	pClone->i_CommandID = i_CommandID;
	pClone->i_MenuItemType = i_MenuItemType;
	pClone->i_MenuItemState = i_MenuItemState;

	pClone->h_Wnd = h_Wnd;
	pClone->h_ParentWnd = h_ParentWnd;

	pClone->p_Parent = _pParent;
	pClone->p_ActivationWindow = p_ActivationWindow;
	pClone->p_MainWindow = p_MainWindow;

	IdToUIObject_map_t::iterator	ite;
	for (ite = map_IdToChildren.begin ();
		ite != map_IdToChildren.end ();
		++ ite)
	{
		UIObject* pCloneChild = ite->second->Clone (pClone);
		pClone->map_IdToChildren.insert (make_pair (ite->first, pCloneChild));
	}
	for (ite = map_CommandIdToMenuItems.begin ();
		ite != map_CommandIdToMenuItems.end ();
		++ ite)
	{
		UIObject* pCloneChild = ite->second->Clone (pClone);
		pClone->map_CommandIdToMenuItems.insert (make_pair (ite->first, pCloneChild));
	}

	if (NULL != _pParent)
		pClone->p_MainWindow = _pParent->p_MainWindow;
	else
		pClone->p_MainWindow = pClone;
	pClone->rct_Body = rct_Body;
	pClone->b_IsVisible = b_IsVisible;
	pClone->b_IsSelfVisible = b_IsSelfVisible;
	pClone->b_HasFocus = b_HasFocus;
	pClone->b_IsActive = b_IsActive;
	pClone->b_IsForeground = b_IsForeground;
	pClone->b_IsEnabledForInput = b_IsEnabledForInput;
	pClone->b_IsChildWnd = b_IsChildWnd;
	pClone->dw_Style = dw_Style;
	pClone->dw_ExStyle = dw_ExStyle;
	pClone->ui_ShowState = ui_ShowState;

	pClone->h_Menu = h_Menu;

	return pClone;
}


//
bool UIObject::Diff (UIObject& _rObject, bool _bPreserveLastDiff)
{
	unsigned int uiStart = o_Diff.ui_Difference;
	if (false == _bPreserveLastDiff)
		o_Diff.ui_Difference = 0;

	if (s_Name != _rObject.s_Name)
	{
		map_IdToOldName.insert (make_pair (s_ID, _rObject.s_Name));
		//o_Diff.bf_Flags.b_Name = true;
	}
	if (i_MenuItemState != _rObject.i_MenuItemState)
		o_Diff.bf_Flags.b_MenuItemState = true;
	if (rct_Body != _rObject.rct_Body)
		o_Diff.bf_Flags.b_RectBody = true;
	if (b_IsVisible != _rObject.b_IsVisible)
		o_Diff.bf_Flags.b_IsVisible = true;
	if (b_HasFocus != _rObject.b_HasFocus)
		o_Diff.bf_Flags.b_HasFocus = true;
	if (b_IsActive != _rObject.b_IsActive)
		o_Diff.bf_Flags.b_IsActive = true;
	if (b_IsForeground != _rObject.b_IsForeground)
		o_Diff.bf_Flags.b_IsForeground = true;
	if (b_IsEnabledForInput != _rObject.b_IsEnabledForInput)
		o_Diff.bf_Flags.b_IsEnabledForInput = true;
	if (dw_Style != _rObject.dw_Style)
		o_Diff.bf_Flags.b_Style = true;
	if (dw_ExStyle != _rObject.dw_ExStyle)
		o_Diff.bf_Flags.b_ExStyle = true;
	if (ui_ShowState != _rObject.ui_ShowState)
		o_Diff.bf_Flags.b_ShowState = true;

	{
		IdToUIObject_map_t::iterator	iteThis;
		for (iteThis = map_IdToChildren.begin ();
			iteThis != map_IdToChildren.end ();
			++ iteThis)
		{
			UIObject* pThisChild = iteThis->second;

			IdToUIObject_map_t::iterator	iteOther;
			iteOther = _rObject.map_IdToChildren.find (iteThis->first);
			if (iteOther == _rObject.map_IdToChildren.end ())
			{
				pThisChild->o_Diff.bf_Flags.b_New = true;
				o_Diff.bf_Flags.b_ChildIsDifferent = true;

				if (true == pThisChild->b_IsVisible)
					pThisChild->o_Diff.bf_Flags.b_IsVisible = true;
				if (true == pThisChild->b_HasFocus)
					pThisChild->o_Diff.bf_Flags.b_HasFocus = true;
				if (true == pThisChild->b_IsActive)
					pThisChild->o_Diff.bf_Flags.b_IsActive = true;
				if (true == pThisChild->b_IsForeground)
					pThisChild->o_Diff.bf_Flags.b_IsForeground = true;
				if (true == pThisChild->b_IsEnabledForInput)
					pThisChild->o_Diff.bf_Flags.b_IsEnabledForInput = true;
			}
			else
				o_Diff.bf_Flags.b_ChildIsDifferent |= pThisChild->Diff (*iteOther->second, _bPreserveLastDiff);
		}
	}
	{
		IdToUIObject_map_t::iterator	iteOther;
		for (iteOther = _rObject.map_IdToChildren.begin ();
			iteOther != _rObject.map_IdToChildren.end ();
			++ iteOther)
		{
			UIObject* pOtherChild = iteOther->second;
			if (true == pOtherChild->o_Diff.bf_Flags.b_Missing)
				continue;

			IdToUIObject_map_t::iterator	iteThis;
			iteThis = map_IdToChildren.find (iteOther->first);
			if (iteThis == map_IdToChildren.end ())
			{
				UIObject* pThisChild = pOtherChild->Clone (this);
				pThisChild->o_Diff.bf_Flags.b_Missing = true;
				map_IdToChildren.insert (make_pair (pThisChild->s_ID, pThisChild));

				o_Diff.bf_Flags.b_ChildIsDifferent = true;
			}
		}
	}

	return (0 != o_Diff.ui_Difference);
}


//
bool UIObject::Rollback (UIObject* _pOriginal, String& _rError)
{
	if (0 == o_Diff.ui_Difference)
		return true;

	if (true == o_Diff.bf_Flags.b_New)
	{
		if ((UIOBJECT__WINDOW != c_ObjectType) &&
			(UIOBJECT__DIALOG != c_ObjectType) &&
			(UIOBJECT__POPUP_MENU != c_ObjectType))
			return true;
		TRACE ("closing : [%s] %s - [%s] %s\n", 
				(const char*)p_ActivationWindow->s_ClassName, (const char*)p_ActivationWindow->s_Name,
				(const char*)s_ClassName, (const char*)s_Name);
		PostMessage (p_ActivationWindow->h_Wnd, WM_SYSCOMMAND, SC_CLOSE, 0);
		Sleep (1);
		PostMessage (p_ActivationWindow->h_Wnd, WM_CLOSE, 0, 0);
		Sleep (1);
		return true;
	}
	if (true == b_StartMenu)
	{
		// the start menu needs to be handled differently from other windows ...
		if (false == o_Diff.bf_Flags.b_IsVisible)
			return true;

		INPUT input [2];
		input [0].type = INPUT_KEYBOARD;
		input [0].ki.wScan = 0;
		input [0].ki.wVk = VK_LWIN;
		input [0].ki.dwFlags = KEYEVENTF_EXTENDEDKEY;
		input [1].type = INPUT_KEYBOARD;
		input [1].ki.wScan = 0;
		input [1].ki.wVk = VK_LWIN;
		input [1].ki.dwFlags = KEYEVENTF_KEYUP|KEYEVENTF_EXTENDEDKEY;
		SendInput (2, input, sizeof (INPUT));
		return true;
	}

	bool bComplete = true;

	//
	if (true == o_Diff.bf_Flags.b_Name)
		SetWindowText (h_Wnd, _pOriginal->s_Name);

	if (true == o_Diff.bf_Flags.b_MenuItemState)
	{
		_rError << "\tcan't rollback menu state : [" << s_Name << "] (" << s_ClassName << ")\n";
		bComplete = false;
	}

	if (true == o_Diff.bf_Flags.b_RectBody)
	{
		// if the window was invisible or not in a displayed show-state, 
		// it's rectangle could be invalid, so we shouldn't set it...
		bool bSetRect = true;
		if (true == _pOriginal->b_IsChildWnd)
			bSetRect = false;
		if (false == _pOriginal->b_IsVisible)
			bSetRect = false;
		if ((SW_HIDE == _pOriginal->p_ActivationWindow->ui_ShowState) &&
			(SW_MINIMIZE == _pOriginal->p_ActivationWindow->ui_ShowState) &&
			(SW_SHOWMINIMIZED == _pOriginal->p_ActivationWindow->ui_ShowState) &&
			(SW_SHOWMINNOACTIVE == _pOriginal->p_ActivationWindow->ui_ShowState))
			bSetRect = false;

		if (true == bSetRect)
			MoveWindow (h_Wnd, 
						_pOriginal->rct_Body.left, 
						_pOriginal->rct_Body.top, 
						_pOriginal->rct_Body.Width (), 
						_pOriginal->rct_Body.Height (),
						TRUE);
	}
	if (true == o_Diff.bf_Flags.b_IsVisible)
	{
		long dwStyle = GetWindowLong (h_Wnd, GWL_STYLE);
		if (true == _pOriginal->b_IsSelfVisible)
			dwStyle |= WS_VISIBLE;
		else
			dwStyle &= ~WS_VISIBLE;
		SetWindowLong (h_Wnd, GWL_STYLE, dwStyle);
	}
	if (true == o_Diff.bf_Flags.b_HasFocus)
	{
		if (true == _pOriginal->b_HasFocus)
			SetFocus (h_Wnd);
	}
	if (true == o_Diff.bf_Flags.b_IsActive)
	{
		if (true == _pOriginal->b_IsActive)
			SetActiveWindow (h_Wnd);
	}
	if (true == o_Diff.bf_Flags.b_IsForeground)
	{
		if (true == _pOriginal->b_IsForeground)
			SetForegroundWindow (h_Wnd);
	}
	if (true == o_Diff.bf_Flags.b_IsEnabledForInput)
		EnableWindow (h_Wnd, (true == _pOriginal->b_IsEnabledForInput)? TRUE : FALSE);

	if (true == o_Diff.bf_Flags.b_Style)
		SetWindowLong (h_Wnd, GWL_STYLE, _pOriginal->dw_Style);

	if (true == o_Diff.bf_Flags.b_ExStyle)
		SetWindowLong (h_Wnd, GWL_EXSTYLE, _pOriginal->dw_ExStyle);

	if (true == o_Diff.bf_Flags.b_ShowState)
		ShowWindow (h_Wnd, _pOriginal->ui_ShowState);


	//
	if (true == o_Diff.bf_Flags.b_ChildIsDifferent)
	{
		IdToUIObject_map_t::iterator	ite;
		for (ite = map_IdToChildren.begin ();
			ite != map_IdToChildren.end ();
			++ ite)
		{
			UIObject* pChild = ite->second;
			if (0 == pChild->o_Diff.ui_Difference)
				continue;
			if (true == pChild->o_Diff.bf_Flags.b_Missing)
			{
				_rError << "\tmissing window : [" << pChild->s_Name << "] (" << pChild->s_ClassName << ")\n";
				bComplete = false;
				continue;
			}

			UIObject* pOriginalChild = NULL;
			IdToUIObject_map_t::iterator	iteOriginal;
			iteOriginal = _pOriginal->map_IdToChildren.find (pChild->s_ID);
			if (iteOriginal != _pOriginal->map_IdToChildren.end ())
				pOriginalChild = iteOriginal->second;
			bComplete &= pChild->Rollback (pOriginalChild, _rError);
		}
	}
	return bComplete;
}


//
bool UIObject::RollbackStartMenu (UIObject* _pOriginal, String& _rError)
{
	if (0 == o_Diff.ui_Difference)
		return true;

	if (true == b_StartMenu)
	{
		if (false == o_Diff.bf_Flags.b_IsVisible)
			return true;

		INPUT input [2];
		input [0].type = INPUT_KEYBOARD;
		input [0].ki.wScan = 0;
		input [0].ki.wVk = VK_LWIN;
		input [0].ki.dwFlags = KEYEVENTF_EXTENDEDKEY;
		input [1].type = INPUT_KEYBOARD;
		input [1].ki.wScan = 0;
		input [1].ki.wVk = VK_LWIN;
		input [1].ki.dwFlags = KEYEVENTF_KEYUP|KEYEVENTF_EXTENDEDKEY;
		SendInput (2, input, sizeof (INPUT));
		return true;
	}

	//
	if (true == o_Diff.bf_Flags.b_ChildIsDifferent)
	{
		IdToUIObject_map_t::iterator	ite;
		for (ite = map_IdToChildren.begin ();
			ite != map_IdToChildren.end ();
			++ ite)
		{
			UIObject* pChild = ite->second;
			if (0 == pChild->o_Diff.ui_Difference)
				continue;
			if (true == pChild->o_Diff.bf_Flags.b_Missing)
				continue;

			UIObject* pOriginalChild = NULL;
			IdToUIObject_map_t::iterator	iteOriginal;
			iteOriginal = _pOriginal->map_IdToChildren.find (pChild->s_ID);
			if (iteOriginal != _pOriginal->map_IdToChildren.end ())
				pOriginalChild = iteOriginal->second;
			pChild->RollbackStartMenu (pOriginalChild, _rError);
		}
	}
	return true;
}


//
String UIObject::GetShowState (void)
{
	if (SW_HIDE == ui_ShowState)
		return "sw_hide";
	if (SW_SHOWNORMAL == ui_ShowState)
		return "sw_show";
	if (SW_NORMAL == ui_ShowState)
		return "sw_show";
	if (SW_SHOWMINIMIZED == ui_ShowState)
		return "sw_min";
	if (SW_MINIMIZE == ui_ShowState)
		return "sw_min";
	if (SW_SHOWMAXIMIZED == ui_ShowState)
		return "sw_show";
	if (SW_MAXIMIZE == ui_ShowState)
		return "sw_show";
	if (SW_SHOWNOACTIVATE == ui_ShowState)
		return "sw_show";
	if (SW_SHOW == ui_ShowState)
		return "sw_show";
	if (SW_SHOWMINNOACTIVE == ui_ShowState)
		return "sw_min";
	if (SW_SHOWNA == ui_ShowState)
		return "sw_show";
	if (SW_RESTORE == ui_ShowState)
		return "sw_show";
	if (SW_SHOWDEFAULT == ui_ShowState)
		return "sw_show";
	if (SW_FORCEMINIMIZE == ui_ShowState)
		return "sw_min";
	if (SW_MAX == ui_ShowState)
		return "sw_show";

	String s;
	// s << "show state : " << (long) ui_ShowState;
	s << (long) ui_ShowState;
	return s;
}


//
void UIObject::GetNeighbouringSibling (UIObject* _pObject, UIObject_dq_t& _rdqObjects)
{
	IdToUIObject_map_t::iterator	ite;
	for (ite = map_IdToChildren.begin ();
		ite != map_IdToChildren.end ();
		++ ite)
	{
		UIObject* pSibling = ite->second;
		if (pSibling == _pObject)
			continue;
		if ("Static" == pSibling->s_ClassName)
			continue;
		bool bIsNeighbour = false;

		// check for sibling on left
		if ((_pObject->rct_Body.right + 50 > pSibling->rct_Body.left) &&
			(_pObject->rct_Body.right < pSibling->rct_Body.left))
		{
			int iOffset = _pObject->rct_Body.CenterPoint ().y - pSibling->rct_Body.CenterPoint ().y;
			if ((iOffset > -5) && (iOffset < 5))
				bIsNeighbour = true;
		}
		// check for sibling below
		if ((_pObject->rct_Body.bottom + 25 > pSibling->rct_Body.top) &&
			(_pObject->rct_Body.bottom < pSibling->rct_Body.top))
		{
			int iOffset = _pObject->rct_Body.left - pSibling->rct_Body.left;
			if ((iOffset > -5) && (iOffset < 5))
				bIsNeighbour = true;
		}

		if (false == bIsNeighbour)
			continue;

		// check children
		if (("Edit" != pSibling->s_ClassName) &&
			("ListBox" != pSibling->s_ClassName) &&
			("ComboBox" != pSibling->s_ClassName) &&
			("Button" != pSibling->s_ClassName) &&
			("radiobutton" != pSibling->s_ClassName) &&
			("checkbox" != pSibling->s_ClassName) &&
			("SysListView32" != pSibling->s_ClassName) &&
			("SysTreeView32" != pSibling->s_ClassName))
		{
			pSibling->GetNeighbouringSibling (_pObject, _rdqObjects);
			continue;
		}
		else
		{
			UIObject* pTemp = pSibling->Clone (_pObject->p_Parent);
			pTemp->s_ID = _pObject->s_ID;
			pTemp->s_Name = _pObject->s_Name;
			UIObject::sdq_TemporaryObjects.push_back (pTemp);

			_rdqObjects.push_back (pTemp);
		}
	}
}


//
void UIObject::GetNeighbouringSibling (UIObject_dq_t& _rdqObjects)
{
	if (NULL == p_Parent)
		return;
	p_Parent->GetNeighbouringSibling (this, _rdqObjects);
}


//
void UIObject::GetEnclosedSibling (UIObject* _pObject, UIObject_dq_t& _rdqObjects)
{
	IdToUIObject_map_t::iterator	ite;
	for (ite = map_IdToChildren.begin ();
		ite != map_IdToChildren.end ();
		++ ite)
	{
		UIObject* pSibling = ite->second;
		if (pSibling == _pObject)
			continue;
		if (("Static" == pSibling->s_ClassName) || ("groupbox" == pSibling->s_ClassName))
			continue;
		bool bIsNeighbour = false;

		// check for enclosed sibling
		if (_pObject->rct_Body.left > pSibling->rct_Body.left)
			continue;
		if (_pObject->rct_Body.right < pSibling->rct_Body.right)
			continue;
		if (_pObject->rct_Body.top > pSibling->rct_Body.top)
			continue;
		if (_pObject->rct_Body.bottom < pSibling->rct_Body.bottom)
			continue;

		// check children
		if (("Edit" != pSibling->s_ClassName) &&
			("ListBox" != pSibling->s_ClassName) &&
			("ComboBox" != pSibling->s_ClassName) &&
			("Button" != pSibling->s_ClassName) &&
			("radiobutton" != pSibling->s_ClassName) &&
			("checkbox" != pSibling->s_ClassName) &&
			("SysListView32" != pSibling->s_ClassName) &&
			("SysTreeView32" != pSibling->s_ClassName))
		{
			pSibling->GetEnclosedSibling (_pObject, _rdqObjects);
			continue;
		}
		else
			_rdqObjects.push_back (pSibling);
	}
}


//
void UIObject::GetEnclosedSibling (UIObject_dq_t& _rdqObjects)
{
	if (NULL == p_Parent)
		return;
	p_Parent->GetEnclosedSibling (this, _rdqObjects);
}


//
String UIObject::GetStandardizedClassName (void)
{
	if ("ToolbarButton" == s_ClassName)
		return "Button";
	if ("TabCtrlButton" == s_ClassName)
		return "Tab";
	if ("ListViewButton" == s_ClassName)
		return "Item";
	if ("TreeViewButton" == s_ClassName)
		return "TreeItem";
	if ("ListBoxItem" == s_ClassName)
		return "Item";
	if ("ComboBoxItem" == s_ClassName)
		return "Item";
	if ("Dialog Box" == s_ClassName)
		return "Dialog";
	if ("groupbox" == s_ClassName)
		return "Static";
	return s_ClassName;
}
