#pragma once
//#define _CRTDBG_MAP_ALLOC
#include <deque>
#include <map>
#include <set>
#include "nlp_string.h"
using namespace std;


//
#define UIOBJECT__UNKNOWN			(char)'u'
#define UIOBJECT__WINDOW			(char)'w'
#define UIOBJECT__DIALOG			(char)'d'
#define UIOBJECT__TOOLBAR_BUTTON	(char)'t'
#define UIOBJECT__LISTVIEW_BUTTON	(char)'l'
#define UIOBJECT__MENU_ITEM			(char)'m'
#define UIOBJECT__TABCTRL_BUTTON	(char)'c'
#define UIOBJECT__POPUP_MENU		(char)'p'
#define UIOBJECT__LISTBOX_ITEM		(char)'x'
#define UIOBJECT__COMBOBOX_ITEM		(char)'b'
#define UIOBJECT__TREEVIEW_ITEM		(char)'r'

//
class Request;
class UIObject;
typedef deque <UIObject*>				UIObject_dq_t;
typedef multimap <String, UIObject*>	NameToUIObject_mmap_t;
typedef map <String, UIObject*>			IdToUIObject_map_t;
typedef map <HWND, UIObject*>			HWndToUIObject_map_t;
typedef map <String, long>				WindowClassToCount_map_t;
typedef map <String, String>			String_map_t;
typedef deque <Request*>				Request_dq_t;
typedef set <DWORD>						ProcessID_set_t;


//
class Request
{
	public:
		static	Request_dq_t	sdq_AllRequests;

		String	s_RawText;
		char	c_Type;
		String	s_UIObjectID;
		String	s_Action;
		String	s_Data;
		String	s_Error;

		String_dq_t	dq_AffectedUIObjectIDs;

		Request (void);
		~Request (void);

		static void Clear (void);
};


//
class SystemConfiguration
{
	public:
		UIObject_dq_t			dq_AllUIObjects;
		NameToUIObject_mmap_t	mmap_NameToUIObject;
		IdToUIObject_map_t		map_IdToUIObject;
		HWndToUIObject_map_t	map_HWndToUIObject;
		IdToUIObject_map_t		map_IdToTopmostUIObjects;
		UIObject*				p_ForegroundObject;
		UIObject*				p_FocusedObject;

		SystemConfiguration (void);
		~SystemConfiguration (void);

		bool IsChanging (SystemConfiguration& _rCurrent, ProcessID_set_t& _rsetProcessesToIgnore);
		bool Diff (SystemConfiguration& _rCurrent, bool _bPreserveLastDiff);
		bool Rollback (SystemConfiguration& _rOriginal, String& _rError); 
		bool RollbackStartMenu (SystemConfiguration& _rOriginal, String& _rError); 

		void ListAllObjects (UIObject_dq_t* _pdqUIObjects);
		bool Find (const char* _zName, UIObject_dq_t* _pdqUIObjects);
		bool FindApproximate (const char* _zLowercaseName, UIObject_dq_t* _pdqUIObjects);
		UIObject* Find (const char* _zID);
		UIObject* Find (HWND _hWnd);
		void CreateWindowTree (void);
};


//
#define OBJECT_HAS_CHANGED	0xFFFE

struct UIObjectChangeFlags
{
	bool	b_ChildIsDifferent	: 1;

	bool	b_Name				: 1;
	bool	b_MenuItemState		: 1;
	bool	b_RectBody			: 1;
	bool	b_IsVisible			: 1;
	bool	b_HasFocus			: 1;
	bool	b_IsActive			: 1;
	bool	b_IsForeground		: 1;
	bool	b_IsEnabledForInput	: 1;
	bool	b_Style				: 1;
	bool	b_ExStyle			: 1;
	bool	b_ShowState			: 1;
	bool	b_New				: 1;
	bool	b_Missing			: 1;
};


//
union UIObjectChange
{
	UIObjectChangeFlags		bf_Flags;
	unsigned int			ui_Difference;
};


String& operator<< (String& _rLeft, const UIObjectChange& _oDiff);


//
class UIObject
{
	public:
		static SystemConfiguration*			p_CurrentSystemConfig;
		static WindowClassToCount_map_t		map_WindowClassToCount;
		static DWORD						dw_MyProcessID;
		static ProcessID_set_t				set_ProcessIDs;
		static String_set_t					set_ClassesToIgnore;
		static String_set_t					set_WindowsToIgnore;
		static String_map_t					map_IdToOldName;
		static bool							b_StartButtonIdentified;
		static bool							b_StartMenuIdentified;
		static long							sl_ObjectCount;
		static UIObject_dq_t				sdq_TemporaryObjects;

		UIObjectChange	o_Diff;

		char			c_ObjectType;
		String			s_ID;
		String			s_Name;
		String			s_LowercaseName;
		String			s_ExtendedName;
		String			s_ClassName;
		UIObject*		p_MainWindow;
		UIObject*		p_ActivationWindow;
		DWORD			dw_ProcessID;
		long			i_CommandID;
		UINT			i_MenuItemType;
		UINT			i_MenuItemState;
		bool			b_StartMenu;

		HWND			h_Wnd;
		HWND			h_ParentWnd;

		UIObject*			p_Parent;
		IdToUIObject_map_t	map_IdToChildren;
		IdToUIObject_map_t	map_CommandIdToMenuItems;

		CRect			rct_Body;
		bool			b_IsVisible;
		bool			b_IsSelfVisible;
		bool			b_HasFocus;
		bool			b_IsActive;
		bool			b_IsForeground;
		bool			b_IsEnabledForInput;
		bool			b_IsChildWnd;
		DWORD			dw_Style;
		DWORD			dw_ExStyle;
		UINT			ui_ShowState;

		HMENU			h_Menu;

		// may not work ...
		// CScrollBar*		p_ScrollBarHorz;
		// CScrollBar*		p_ScrollBarVert;
		// SCROLLINFO*		p_ScrollInfoHorz;
		// SCROLLINFO*		p_ScrollInfoVert;


		void Init (void);
		UIObject (void);
		UIObject (CWnd* _pWnd);
		UIObject (UIObject* _pParentWnd, HWND _hToolbarWnd, int _iButtonIndex, char _cObjectType);
		UIObject (UIObject* _pParentWnd, HWND _hTreeWnd, HTREEITEM _hItem, char _cObjectType);
		UIObject (UIObject* _pParentWnd, CMenu* _pMenu, int _iItemIndex);
		~UIObject (void);

		bool ExtractWindowInformation (CWnd* _pWnd);
		void ExtractToolbarInformation (UIObject* _pUIObject);
		void ExtractListViewInformation (UIObject* _pUIObject);
		void ExtractTabControlInformation (UIObject* _pUIObject);
		void ExtractListBoxInformation (UIObject* _pUIObject);
		void ExtractComboBoxInformation (UIObject* _pUIObject);
		void ExtractTreeViewInformation (void);

		void ExtractMenuInformation (UIObject* _pParentWnd, CMenu* _pMenu, int _iItemIndex);
		void ExtractToolbarButtonInformation (HWND _hToolbarWnd, int _iButtonIndex);
		void ExtractListViewButtonInformation (HWND _hToolbarWnd, int _iButtonIndex);
		void ExtractTabCtrlButtonInformation (HWND _hToolbarWnd, int _iButtonIndex);
		void ExtractListBoxItemInformation (HWND _hListBoxWnd, int _iItemIndex);
		void ExtractComboBoxItemInformation (HWND _hComboBoxWnd, int _iItemIndex);
		void ExtractTreeViewItemInformation (HWND _hTreeView, HTREEITEM _hItem);
		void ExtractTreeViewChildren (HWND _hTreeView, HTREEITEM _hItem);

		void AddChild (UIObject* _pUIObject);

		String GetShowState (void);
		String GetStandardizedClassName (void);
		void GetNeighbouringSibling (UIObject_dq_t& _rdqObjects);
		void GetNeighbouringSibling (UIObject* _pObject, UIObject_dq_t& _rdqObjects);
		void GetEnclosedSibling (UIObject_dq_t& _rdqObjects);
		void GetEnclosedSibling (UIObject* _pObject, UIObject_dq_t& _rdqObjects);
		UIObject* Clone (UIObject* _pParent);
		bool Diff (UIObject& _rObject, bool _bPreserveLastDiff);
		bool Rollback (UIObject* _pObject, String& _rError);
		bool RollbackStartMenu (UIObject* _pObject, String& _rError);

		static void LoadIgnoreList (void);
		static void ScanFromWindow (CWnd* _pWnd, ProcessID_set_t* _psetProcessesToIgnore = NULL);
		static void Clear (void);
		static void ClearTemporaryObjects (void);
};
