#include "stdafx.h"
#include "Identify.h"

#include "IdentifyDoc.h"
#include "WindowDiffView.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif


//
IMPLEMENT_DYNCREATE(WindowDiffView, IdentifyView)
BEGIN_MESSAGE_MAP(WindowDiffView, IdentifyView)
	ON_COMMAND(ID_VIEW_SHOWNAMELESSWINDOWS, &WindowDiffView::OnViewShownamelesswindows)
	ON_UPDATE_COMMAND_UI(ID_VIEW_SHOWNAMELESSWINDOWS, &WindowDiffView::OnUpdateViewShownamelesswindows)
END_MESSAGE_MAP()


//
WindowDiffView::WindowDiffView()
{
	b_ShowNamelessWindows = false;
	h_SelectedItem = NULL;
}

WindowDiffView::~WindowDiffView()
{
}


//
void WindowDiffView::OnUpdate (CView* /*pSender*/, LPARAM _lHint, CObject* /*pHint*/)
{
	if (UPDATE_WINDOW_VIEWS != _lHint)
		return;
	PopulateWindowDiff ();
}


//
void WindowDiffView::PopulateWindowDiff (void)
{
	IdentifyDoc* pDoc = GetDocument ();
	CTreeCtrl& rTreeCtrl = IdentifyView::GetTreeCtrl ();
	rTreeCtrl.SetRedraw (FALSE);
	rTreeCtrl.DeleteAllItems ();

	IdToUIObject_map_t::iterator	ite;
	for (ite = pDoc->p_CurrentConfiguration->map_IdToTopmostUIObjects.begin ();
		ite != pDoc->p_CurrentConfiguration->map_IdToTopmostUIObjects.end ();
		++ ite)
	{
		UIObject* pUIObject = ite->second;
		if (0 == pUIObject->o_Diff.ui_Difference)
			continue;
		InsertUIObjectToTree (rTreeCtrl, pUIObject);
	}

	rTreeCtrl.SortChildren (NULL);
	rTreeCtrl.SetRedraw (TRUE);
	rTreeCtrl.RedrawWindow ();
}


//
void WindowDiffView::InsertUIObjectToTree (CTreeCtrl& _rTreeCtrl, UIObject* _pUIObject, HTREEITEM _hParent)
{
	if (0 == _pUIObject->o_Diff.ui_Difference)
		return;


	HTREEITEM hItem = NULL;
	if (0 != (OBJECT_HAS_CHANGED & _pUIObject->o_Diff.ui_Difference))
	{
		String sName;
		sName += _pUIObject->s_Name + " [" + _pUIObject->s_ClassName + "]";

		int iImageIndex = UIObjectToImageIndex (_pUIObject);
		if (0 == iImageIndex)
			iImageIndex = 1;
		if (("ToolbarButton" == _pUIObject->s_ClassName) &&
			(_pUIObject->rct_Body.IsRectEmpty ()))
			++ iImageIndex;

		hItem = _rTreeCtrl.InsertItem (sName, iImageIndex, iImageIndex, _hParent, TVI_LAST);
		_rTreeCtrl.SetItemData (hItem, (DWORD_PTR)_pUIObject);

		HTREEITEM hProperties = _rTreeCtrl.InsertItem ("Properties", hItem, TVI_LAST);
		InsertWindowInformation (_rTreeCtrl, _pUIObject, hProperties);

		if (false == _pUIObject->map_CommandIdToMenuItems.empty ())
		{
			HTREEITEM hItemMenu = _rTreeCtrl.InsertItem ("Menu", hItem, TVI_LAST);
			IdToUIObject_map_t::iterator	iteMenu;
			for (iteMenu = _pUIObject->map_CommandIdToMenuItems.begin ();
				iteMenu != _pUIObject->map_CommandIdToMenuItems.end ();
				++ iteMenu)
				InsertMenuInformation (_rTreeCtrl, iteMenu->second, hItemMenu);
		}
	}
	else
	{
		String sName;
		sName += _pUIObject->s_Name + " [" + _pUIObject->s_ClassName + "]";
		hItem = _rTreeCtrl.InsertItem (sName, 0, 0, _hParent, TVI_LAST);
	}		

	IdToUIObject_map_t::iterator	ite;
	for (ite = _pUIObject->map_IdToChildren.begin ();
		ite != _pUIObject->map_IdToChildren.end ();
		++ ite)
	{
		UIObject* pChild = ite->second;
		if (0 == pChild->o_Diff.ui_Difference)
			continue;
		InsertUIObjectToTree (_rTreeCtrl, ite->second, hItem);
	}
}


//
void WindowDiffView::OnViewShownamelesswindows()
{
	if (true == b_ShowNamelessWindows)
		b_ShowNamelessWindows = false;
	else
		b_ShowNamelessWindows = true;

	PopulateWindowDiff ();
}


//
void WindowDiffView::OnUpdateViewShownamelesswindows(CCmdUI *pCmdUI)
{
	pCmdUI->SetCheck ((true == b_ShowNamelessWindows)? TRUE : FALSE);
}






// WindowDiffView diagnostics

#ifdef _DEBUG
void WindowDiffView::AssertValid() const
{
	IdentifyView::AssertValid();
}

void WindowDiffView::Dump(CDumpContext& dc) const
{
	IdentifyView::Dump(dc);
}

IdentifyDoc* WindowDiffView::GetDocument() const // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(IdentifyDoc)));
	return (IdentifyDoc*)m_pDocument;
}
#endif //_DEBUG







