#include "stdafx.h"
#include "Identify.h"

#include "IdentifyDoc.h"
#include "WindowListView.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif


//
IMPLEMENT_DYNCREATE(WindowListView, IdentifyView)
BEGIN_MESSAGE_MAP(WindowListView, IdentifyView)
	ON_COMMAND(ID_VIEW_SHOWNAMELESSWINDOWS, &WindowListView::OnViewShownamelesswindows)
	ON_UPDATE_COMMAND_UI(ID_VIEW_SHOWNAMELESSWINDOWS, &WindowListView::OnUpdateViewShownamelesswindows)
	ON_COMMAND(ID_VIEW_SHOWTOOLTIPWINDOWS, &WindowListView::OnViewShowtooltipwindows)
	ON_UPDATE_COMMAND_UI(ID_VIEW_SHOWTOOLTIPWINDOWS, &WindowListView::OnUpdateViewShowtooltipwindows)
END_MESSAGE_MAP()


//
WindowListView::WindowListView()
{
	b_ShowNamelessWindows = true;
	b_ShowTooltips = false;
	h_SelectedItem = NULL;
}

WindowListView::~WindowListView()
{
}


//
void WindowListView::OnUpdate (CView* /*pSender*/, LPARAM _lHint, CObject* /*pHint*/)
{
	if (UPDATE_WINDOW_VIEWS != _lHint)
		return;
	PopulateWindowList ();
}


//
void WindowListView::PopulateWindowList (void)
{
	IdentifyDoc* pDoc = GetDocument ();
	CTreeCtrl& rTreeCtrl = IdentifyView::GetTreeCtrl ();
	rTreeCtrl.SetRedraw (FALSE);
	rTreeCtrl.DeleteAllItems ();

	IdToUIObject_map_t::iterator	ite;
	for (ite = pDoc->p_CurrentConfiguration->map_IdToTopmostUIObjects.begin ();
		ite != pDoc->p_CurrentConfiguration->map_IdToTopmostUIObjects.end ();
		++ ite)
	{
		UIObject* pUIObject = ite->second;
		if ((false == b_ShowNamelessWindows) && ("" == pUIObject->s_Name))
			continue;
		if ((false == b_ShowTooltips) && 
			(
				("tooltips_class32" == pUIObject->s_ClassName) ||
				("VsTipWindow" == pUIObject->s_ClassName) ||
				("OfficeTooltip" == pUIObject->s_ClassName) ||
				("VBBubble" == pUIObject->s_ClassName) ||
				("MS_WINNOTE" == pUIObject->s_ClassName)
			))
			continue;
		InsertUIObjectToTree (rTreeCtrl, pUIObject);
	}

	rTreeCtrl.SortChildren (NULL);
	rTreeCtrl.SetRedraw (TRUE);
	rTreeCtrl.RedrawWindow ();
}


//
void WindowListView::InsertUIObjectToTree (CTreeCtrl& _rTreeCtrl, UIObject* _pUIObject, HTREEITEM _hParent)
{
	if (true == _pUIObject->o_Diff.bf_Flags.b_Missing)
		return;

	String sName;
	sName += _pUIObject->s_Name + " [" + _pUIObject->s_ClassName + "]";

	int iImageIndex = UIObjectToImageIndex (_pUIObject);
	if (0 == iImageIndex)
		iImageIndex = 1;
	if (("ToolbarButton" == _pUIObject->s_ClassName) &&
		(_pUIObject->rct_Body.IsRectEmpty ()))
		++ iImageIndex;

	HTREEITEM hItem = _rTreeCtrl.InsertItem (sName, iImageIndex, iImageIndex, _hParent, TVI_LAST);
	_rTreeCtrl.SetItemData (hItem, (DWORD_PTR)_pUIObject);

	HTREEITEM hProperties = _rTreeCtrl.InsertItem ("Properties", hItem, TVI_LAST);
	InsertWindowInformation (_rTreeCtrl, _pUIObject, hProperties);

	if (false == _pUIObject->map_CommandIdToMenuItems.empty ())
	{
		HTREEITEM hItemMenu = _rTreeCtrl.InsertItem ("Menu", hItem, TVI_LAST);
		IdToUIObject_map_t::iterator	iteMenu;
		for (iteMenu = _pUIObject->map_CommandIdToMenuItems.begin ();
			iteMenu != _pUIObject->map_CommandIdToMenuItems.end ();
			++ iteMenu)
			InsertMenuInformation (_rTreeCtrl, iteMenu->second, hItemMenu);
	}

	IdToUIObject_map_t::iterator	ite;
	for (ite = _pUIObject->map_IdToChildren.begin ();
		ite != _pUIObject->map_IdToChildren.end ();
		++ ite)
	{
		InsertUIObjectToTree (_rTreeCtrl, ite->second, hItem);
	}
}


//
void WindowListView::OnViewShownamelesswindows()
{
	if (true == b_ShowNamelessWindows)
		b_ShowNamelessWindows = false;
	else
		b_ShowNamelessWindows = true;

	PopulateWindowList ();
}


//
void WindowListView::OnUpdateViewShownamelesswindows (CCmdUI *pCmdUI)
{
	pCmdUI->SetCheck ((true == b_ShowNamelessWindows)? TRUE : FALSE);
}


//
void WindowListView::OnViewShowtooltipwindows ()
{
	if (true == b_ShowTooltips)
		b_ShowTooltips = false;
	else
		b_ShowTooltips = true;

	PopulateWindowList ();
}


//
void WindowListView::OnUpdateViewShowtooltipwindows (CCmdUI *pCmdUI)
{
	pCmdUI->SetCheck ((true == b_ShowTooltips)? TRUE : FALSE);
}






// WindowListView diagnostics

#ifdef _DEBUG
void WindowListView::AssertValid() const
{
	IdentifyView::AssertValid();
}

void WindowListView::Dump(CDumpContext& dc) const
{
	IdentifyView::Dump(dc);
}

IdentifyDoc* WindowListView::GetDocument() const // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(IdentifyDoc)));
	return (IdentifyDoc*)m_pDocument;
}
#endif //_DEBUG








